# basically variables/functions signatures, but base class for many features
from pyrser.type_system.symbol import *


class Signature(Symbol):
    """
    Describe a function or variable signature for the language
    """
    def __init__(self, name: str):
        super().__init__(name)

    def to_fmt(self):
        raise TypeError("Signature.to_fmt must be redefined")

    def __str__(self):
        return str(self.to_fmt())



# var for type checking (variables)
from pyrser import fmt
from pyrser.type_system.signature import *
from pyrser.type_system.type_name import *


class Var(Signature):
    """
    Describe a variable signature for the language
    """

    def __init__(self, name: str, tret: str):
        super().__init__(name)
        if not isinstance(tret, TypeName):
            tret = TypeName(tret)
        self.tret = tret

    @property
    def is_polymorphic(self) -> bool:
        return self.tret.is_polymorphic

    def to_fmt(self):
        """
        Return an Fmt representation for pretty-printing
        """
        params = ""
        txt = fmt.sep(" ", ['var'])
        name = self.show_name()
        if name != "":
            txt.lsdata.append(name)
        txt.lsdata.append(': ' + self.tret)
        return txt

    def internal_name(self):
        """
        Return the unique internal name
        """
        unq = super().internal_name()
        if self.tret is not None:
            unq += "_" + self.tret
        return unq



from pyrser.parsing.functors import Functor
from pyrser.parsing.functors import PeekChar, Char, PeekText, Text, Range
from pyrser.parsing.functors import UntilChar
from pyrser.parsing.functors import Call, CallTrue
from pyrser.parsing.functors import Complement, LookAhead, Neg, Until
from pyrser.parsing.functors import Hook, Rule
from pyrser.parsing.functors import Directive, DirectiveWrapper
from pyrser.parsing.functors import Decorator, DecoratorWrapper
from pyrser.parsing.functors import Alt, Seq
from pyrser.parsing.functors import Rep0N, Rep1N, RepOptional
from pyrser.parsing.functors import Capture, Scope, Bind, DeclNode
from pyrser.parsing.functors import Error
from pyrser.parsing.base import BasicParser, Parser, MetaBasicParser
from pyrser.parsing.stream import Stream
from pyrser.parsing.node import Node


__all__ = [
    'Alt',
    'BasicParser',
    'Bind',
    'Call',
    'CallTrue',
    'Capture',
    'Char',
    'Complement',
    'DeclNode',
    'Directive',
    'DirectiveWrapper',
    'Decorator',
    'DecoratorWrapper',
    'Error',
    'Functor',
    'Hook',
    'LookAhead',
    'MetaBasicParser',
    'Neg',
    'Node',
    'Parser',
    'PeekChar',
    'PeekText',
    'Range',
    'Rule',
    'Rep0N',
    'Rep1N',
    'RepOptional',
    'Scope',
    'Seq',
    'Stream',
    'Text',
    'Until',
    'UntilChar'
]



# IR allow to represent algorithmic bricks of PEG parsing

class IR:
    """ Base class for Internal Representation of algorithmic bricks. """


class Grammar(IR):
    """Abstraction of a whole grammar."""

    def __init__(self, name: str):
        self.name = name
        self.rules = []

class Rule(IR):
    """Abstraction of a target function. """
    
    def __init__(self, name: str):
        self.name = name
        self.block = None

class Block(IR):
    
    def __init__(self, lsexpr: list):
        if not isinstance(lsexpr, list):
            raise TypeError("Take only a list")
        self.lsexpr = lsexpr

class CallRule(IR):
    pass


class GetC(IR):
    pass

class IncPos(IR):
    pass

class EqualBlock(IR):
    
    def __init__(self, v: str)
        self.value = v
        self.block = None

class RangeBlock(IR):
    pass

class WhileEofBlock(IR):
    """ """
    pass


class ReturnOnEof(IR):
    """If we reach the EOF, we abort parsing directly."""
    pass


class SaveCtx(IR):
    """Temporary save the current parsing context."""
    pass

class RestoreCtx(IR):
    """Restore previous parsing context."""
    pass

class ValidateCtx(IR):
    """Validate current parsing context."""
    pass

class Return(IR):
    """Abstraction of a return statement. return the last boolean state."""
    pass



import sys
from pyrser import meta, parsing


@meta.decorator("trace")
class Trace(parsing.DecoratorWrapper):

    out = None

    def __init__(self, outfile: str=""):
        self.level = 0
        self.indent = 4
        if outfile != "":
            self.out = open(outfile, 'w')
        else:
            self.out = sys.stdout

    def __del__(self):
        self.out.close()

    def begin(self, parser: parsing.BasicParser, pt: parsing.Functor):
        """
        The begin method of the "root" Trace has to open the output
        for all its children tracers.
        """
        item = pt
        while (isinstance(item, parsing.Directive)
               or isinstance(item, parsing.Decorator)):
            item = item.pt

        if isinstance(item, parsing.Rule) is True:
            self.out.write(" " * self.indent * self.level
                           + "[" + item.name + "] Entering\n")
            self.level += 1

        return True

    def end(self,
            result: bool, parser: parsing.BasicParser, pt: parsing.Functor):
        """
        The end method of the "root" Trace has to close the output
        """

        item = pt
        while (isinstance(item, parsing.Directive)
               or isinstance(item, parsing.Decorator)):
            item = item.pt

        if isinstance(item, parsing.Rule) is True:
            self.level -= 1
            rstr = "Failed"
            if result is not False:
                rstr = "Succeeded"
            self.out.write(" " * self.indent * self.level
                           + "[" + item.name + "] " + rstr + "\n")

        return True



# fun for type checking (functions signatures)
from pyrser import fmt
from pyrser.type_system.signature import *
from pyrser.type_system.type_name import *


class Fun(Signature):
    """
    Describe a function signature for the language
    """

    def __init__(self, name: str, tret: str, tparams: list=None, variadic=None):
        if tparams is not None and not isinstance(tparams, list):
            raise TypeError("Fun's parameter list must be a list")
        super().__init__(name)
        if not isinstance(tret, TypeName):
            tret = TypeName(tret)
        self.tret = tret
        self.variadic = variadic
        self.tparams = None
        if tparams is not None and len(tparams) > 0:
            self.tparams = []
            for p in tparams:
                if not isinstance(p, TypeName):
                    p = TypeName(p)
                self.tparams.append(p)

    @property
    def arity(self) -> int:
        return len(self.tparams)

    @property
    def return_type(self) -> str:
        return self.tret

    @property
    def this_type(self) -> str:
        return self.tparams[0]

    @property
    def is_polymorphic(self) -> bool:
        if self.tret.is_polymorphic:
            return True
        for p in self.tparams:
            if p.is_polymorphic:
                return True

    def to_fmt(self):
        """
        Return an Fmt representation for pretty-printing
        """
        params = ""
        txt = fmt.sep(" ", ['fun'])
        name = self.show_name()
        if name != "":
            txt.lsdata.append(name)
        tparams = []
        if self.tparams is not None:
            tparams = list(self.tparams)
        if self.variadic:
            tparams.append('...')
        params = '(' + ", ".join(tparams) + ')'
        txt.lsdata.append(': ' + params)
        txt.lsdata.append('-> ' + self.tret)
        return txt

    def internal_name(self):
        """
        Return the unique internal name
        """
        unq = super().internal_name()
        if self.tparams is not None:
            unq += "_" + "_".join(self.tparams)
        if self.tret is not None:
            unq += "_" + self.tret
        return unq




# This pass is for converting functors into IR algos bricks
# for easy target language transformation
from pyrser import meta
from pyrser import fmt
from pyrser import parsing

@meta.add_method(parsing.Parser)
def to_ir(self) -> parsing.ir.IR:
    gram = parsing.ir.Grammar(self.__class__.__name__)
    for k, v in self.__class__._rules.items():
        if isinstance(v, parsing.Functor):
            rule = parsing.ir.Rule(k)
            rule.expr = v.to_ir()
            gram.rules.append(rule)
    return gram

@meta.add_method(parsing.Rule)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.Hook)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.PeekText)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.PeekChar)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.Text)
def to_ir(self) -> parsing.ir.IR:
    return parsing.ir.Block([
        parsing.ir.ReturnOnEof(),
        parsing.ir.SaveCtx(),
        parsing.ir.GetC(),
        parsing.ir.EqualBlock(self.char).block = parsing.ir.Block([
            parsing.ir.IncPos(),
            parsing.ir.ValidateCtx(),
            parsing.ir.Return()
        ]),
        parsing.ir.RestoreCtx(),
        parsing.ir.Return()
    ])

@meta.add_method(parsing.Char)
def to_ir(self) -> parsing.ir.IR:
    return parsing.ir.Block([
        parsing.ir.ReturnOnEof(),
        parsing.ir.SaveCtx(),
        parsing.ir.GetC(),
        parsing.ir.EqualBlock(self.char).block = parsing.ir.Block([
            parsing.ir.IncPos(),
            parsing.ir.ValidateCtx(),
            parsing.ir.Return()
        ]),
        parsing.ir.RestoreCtx(),
        parsing.ir.Return()
    ])

@meta.add_method(parsing.Scope)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.Directive)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.Capture)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.DeclNode)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.Bind)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.Error)
def to_ir(self) -> parsing.ir.IR:
    pass

# concatenate all IR of a seq
@meta.add_method(parsing.Seq)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.Alt)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.Neg)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.Complement)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.Until)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.UntilChar)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.LookAhead)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.Neg)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.RepOptional)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.Rep0N)
def to_ir(self) -> parsing.ir.IR:
    pass

@meta.add_method(parsing.Rep1N)
def to_ir(self) -> parsing.ir.IR:
    pass



import collections


"""An immutable position in a Stream.

index is the index in the Stream.
lineno is the line number of the position.
col_offset is the column offset of the position on the line.
"""
Position = collections.namedtuple('Position', 'index lineno col_offset')


class Cursor:
    """A mutable position in a Stream.

    It can be initialized or set from an immutable Position.
    """
    def __init__(self, position: Position=Position(0, 1, 1)):
        self._maxindex = self._index = position.index
        self._maxcol = self._col_offset = position.col_offset
        self._maxline = self._lineno = position.lineno
        self._eol = []

    @property
    def index(self) -> int:
        """The current index of the cursor."""
        return self._index

    @property
    def lineno(self) -> int:
        """The current line number of the cursor."""
        return self._lineno

    @property
    def col_offset(self) -> int:
        """The current column offset of the cursor."""
        return self._col_offset

    @property
    def position(self) -> Position:
        """The current position of the cursor."""
        return Position(self._index, self._lineno, self._col_offset)

    @position.setter
    def position(self, position: Position):
        self._index = position.index
        self._lineno = position.lineno
        self._col_offset = position.col_offset

    @property
    def max_readed_position(self) -> Position:
        """The index of the deepest character readed."""
        return Position(self._maxindex, self._maxline, self._maxcol)

    def step_next_char(self):
        """Puts the cursor on the next character."""
        self._index += 1
        self._col_offset += 1
        if self._index > self._maxindex:
            self._maxindex = self._index
            self._maxcol = self._col_offset
            self._maxline = self._lineno

    def step_prev_char(self):
        """Puts the cursor on the previous character."""
        self._col_offset -= 1
        self._index -= 1

    def step_next_line(self):
        """Sets cursor as beginning of next line."""
        self._eol.append(self.position)
        self._lineno += 1
        self._col_offset = 0

    def step_prev_line(self):
        """Sets cursor as end of previous line."""
        #TODO(bps): raise explicit error for unregistered eol
        #assert self._eol[-1].index == self._index
        if len(self._eol) > 0:
            self.position = self._eol.pop()


class Tag:
    """Provide capture facilities"""
    def __init__(self, stream: str, begin: int, end=0):
        self._stream = stream
        self._begin = begin
        if end == 0:
            self._end = begin
        else:
            self._end = end

    def set_begin(self, begin: int):
        self._begin = begin

    def set_end(self, end: int):
        self._end = end

    def __str__(self) -> str:
        if self._begin == self._end:
            return ""
        return self._stream[self._begin:self._end]

    def __repr__(self) -> str:
        return "strid:%d %s:%s" % (id(self._stream), self._begin, self._end)


class Stream:
    """Helps keep track of stream processing progress."""
    def __init__(self, content: str=None, name: str=None):
        self._content = content
        self._len = len(content)
        self._name = name
        self._contexts = []
        self._cursor = Cursor()
        # use to store begin:end => value
        self.value_cache = dict()

    def __len__(self) -> int:
        return self._len

    def __getitem__(self, key: int or slice) -> str:
        return self._content.__getitem__(key)

    @property
    def name(self) -> int:
        """Name of the stream."""
        return self._name

    @property
    def eos_index(self) -> int:
        """End Of Stream index."""
        return self._len

    @property
    def index(self) -> int:
        """The current position index."""
        return self._cursor.position.index

    @property
    def lineno(self) -> int:
        """The current position line number."""
        return self._cursor.lineno

    @property
    def col_offset(self) -> int:
        """The current position column offset."""
        return self._cursor.col_offset

    @property
    def peek_char(self) -> str:
        """The current position character value."""
        return self._content[self._cursor.index]

    @property
    def last_readed_line(self) -> str:
        """Usefull string to compute error message."""
        mpos = self._cursor.max_readed_position
        mindex = mpos.index
        # search last \n
        prevline = mindex - 1 if mindex == self.eos_index else mindex
        while prevline >= 0 and self._content[prevline] != '\n':
            prevline -= 1
        # search next \n
        nextline = mindex
        while nextline < self.eos_index and self._content[nextline] != '\n':
            nextline += 1
        last_line = self._content[prevline + 1:nextline]
        return last_line

    def incpos(self, length: int=1) -> int:
        """Increment the cursor to the next character."""
        if length < 0:
            raise ValueError("length must be positive")
        i = 0
        while (i < length):
            if self._cursor.index < self._len:
                if self.peek_char == '\n':
                    self._cursor.step_next_line()
                self._cursor.step_next_char()
            i += 1
        return self._cursor.index

    def decpos(self, length: int=1) -> int:
        if length < 0:
            raise ValueError("length must be positive")
        if (self._cursor.index - length) < 0:
            raise ValueError("can't go before first byte")
        i = 0
        while (i < length):
            if self.peek_char == '\n':
                self._cursor.step_prev_line()
                i += 1
            else:
                self._cursor.step_prev_char()
            i += 1
        return self._cursor.index

    def save_context(self) -> bool:
        """Save current position."""
        self._contexts.append(self._cursor.position)
        return True

    def restore_context(self) -> bool:
        """Rollback to previous saved position."""
        self._cursor.position = self._contexts.pop()
        return False

    def validate_context(self) -> bool:
        """Discard previous saved position."""
        del self._contexts[-1]
        return True



from pyrser import meta
from pyrser.parsing.base import BasicParser


@meta.hook(BasicParser, "false")
def pred_false(self):
    """
    False in parser. Go to alternative...

    example::

        R = [
            #false | R2
        ]

    """
    return False


@meta.hook(BasicParser, "true")
def pred_true(self):
    """
    True in parser. Do nothing

    example::

        R = [
            R2 | #true
        ]

    """
    return True

@meta.hook(BasicParser, "eq")
def pred_eq(self, n, val):
    """
    Test if a node set with setint or setstr equal a certain value

    example::

        R = [
            __scope__:n
            ['a' #setint(n, 12) | 'b' #setint(n, 14)]
            C
            [#eq(n, 12) D]
        ]

    """
    v1 = n.value
    v2 = val
    if hasattr(val, 'value'):
        v2 = val.value
    if isinstance(v1, int) and not isinstance(v2, int):
        return v1 == int(v2)
    return v1 == v2

@meta.hook(BasicParser, "neq")
def pred_eq(self, n, val):
    """
    Test if a node set with setint or setstr not equal a certain value

    example::

        R = [
            __scope__:n
            ['a' #setint(n, 12) | 'b' #setint(n, 14)]
            C
            [#neq(n, 12) D]
        ]

    """
    v1 = n.value
    v2 = val
    if hasattr(val, 'value'):
        v2 = val.value
    if isinstance(v1, int) and not isinstance(v2, int):
        return v1 != int(v2)
    return v1 != v2



from pyrser.type_system.type_name import TypeName
from pyrser.type_system.symbol import Symbol
from pyrser.type_system.signature import Signature
from pyrser.type_system.evalctx import EvalCtx
from pyrser.type_system.scope import Scope, StateScope
from pyrser.type_system.tuple import Tuple
from pyrser.type_system.type import Type
from pyrser.type_system.var import Var
from pyrser.type_system.val import Val
from pyrser.type_system.fun import Fun
from pyrser.type_system.translator import Translator, MapSourceTranslate, MapTargetTranslate

__all__ = [
    'EvalCtx',
    'Scope',
    'StateScope',
    'Signature',
    'Symbol',
    'Tuple',
    'Type',
    'TypeName',
    'Val',
    'Var',
    'Fun',
    'Translator',
    'MapSourceTranslate',
    'MapTargetTranslate',
]



# val for type checking (literal or ENUM style)
from pyrser import fmt
from pyrser.type_system.signature import *
from pyrser.type_system.type_name import *


class Val(Signature):
    """
    Describe a value signature for the language
    """
    nvalues = 0
    valuniq = dict()

    def __init__(self, value, tret: str):
        if not isinstance(value, str):
            value = str(value)
        self.value = value
        if not isinstance(tret, TypeName):
            tret = TypeName(tret)
        self.tret = tret
        k = self.value + "$" + tret
        idx = 0
        if k not in Val.valuniq:
            Val.nvalues += 1
            Val.valuniq[k] = Val.nvalues
            idx = Val.nvalues
        else:
            idx = Val.valuniq[k]
        super().__init__('$' + str(idx))

    def to_fmt(self):
        """
        Return an Fmt representation for pretty-printing
        """
        params = ""
        txt = fmt.sep(" ", ['val'])
        name = self.show_name()
        if name != "":
            txt.lsdata.append(name)
        txt.lsdata.append('(%s)' % self.value)
        txt.lsdata.append(': ' + self.tret)
        return txt

    def internal_name(self):
        """
        Return the unique internal name
        """
        unq = super().internal_name()
        if self.tret is not None:
            unq += "_" + self.tret
        return unq



# type for type checking
from pyrser import fmt
from pyrser.type_system.scope import *
from pyrser.type_system.type_name import *


class Type(Scope):
    """
    A Type is a named scope similar to ADT (Abstract Data Type).
    So we have member variables (ordered) that represent internal state.
    Member functions (work on an instance of type,
    depend of the language ref/or not).
    Non member variables/functions/values as in a scope.
    """

    def __init__(self, name: str, sig: [Signature]=None):
        super().__init__(TypeName(name))

    def to_fmt(self) -> fmt.indentable:
        """
        Return an Fmt representation for pretty-printing
        """
        qual = "type"
        txt = fmt.sep(" ", [qual])
        txt.lsdata.append(self.show_name())
        if hasattr(self, '_hsig') and len(self._hsig) > 0:
            lsb = []
            for k in sorted(self._hsig.keys()):
                s = self._hsig[k]
                lsb.append(fmt.end("\n", [s.to_fmt()]))
            block = fmt.block(":\n", "", fmt.tab(lsb))
            txt.lsdata.append(block)
        return txt

    @property
    def type_name(self) -> TypeName:
        return self.name



from pyrser import meta
from pyrser.parsing import Node
from pyrser.parsing.base import BasicParser


@meta.hook(BasicParser, "set")
def set_node(self, dst, src):
    """
        Basically copy one node to another.
        usefull to transmit a node from a terminal
        rule as result of the current rule.

        example::

            R = [
                In : node #set(_, node)
            ]

        here the node return by the rule In is
        also the node return by the rule R
    """
    if not isinstance(src, Node):
        dst.value = src
    else:
        dst.set(src)
        idsrc = id(src)
        iddst = id(dst)
        if iddst not in self.id_cache:
            print("DST: %s" % repr(dst))
            print("RULE_NODES %s" % repr(self.rule_nodes))
        if idsrc in self.id_cache:
            k = self.id_cache[idsrc]
            k2 = self.id_cache[iddst]
            if k in self.rule_nodes:
                self.tag_cache[k2] = self.tag_cache[k]
    return True

@meta.hook(BasicParser, "setint")
def set_node_as_int(self, dst, src):
    dst.value = int(src)
    return True

@meta.hook(BasicParser, "setstr")
def set_node_as_str(self, dst, src):
    dst.value = src
    return True

@meta.hook(BasicParser, "get")
def get_subnode(self, dst, ast, expr):
    dst.value = eval('ast' + expr)
    return True



# This pass is for debug only
from pyrser import meta
from pyrser import parsing


@meta.add_method(parsing.Parser)
def to_dsl(self):
    res = ""
    for k, v in self.__class__._rules.items():
        if isinstance(v, parsing.Functor):
            res += "\n\t{} ::= {}\t;".format(k, v.to_dsl(1))
        else:
            res += "\n\t{} ::= {};".format(k, repr(v))
    return res


def _dumpParam(pt):
    res = ""
    if len(pt.param) > 0:
        lsparam = []
        for p in pt.param:
            if p[1] is str:
                lsparam.append('"' + p[0] + '"')
            else:
                lsparam.append(p[0])
        res = "({})".format(", ".join(lsparam))
    return res


@meta.add_method(parsing.Rule)
def to_dsl(self, level=0):
    return "{}{}".format('\t' * level, self.name)


@meta.add_method(parsing.Hook)
def to_dsl(self, level=0):
    return "{}#{}{}".format('\t' * level, self.name, _dumpParam(self))


@meta.add_method(parsing.Call)
def to_dsl(self, level=0):
    #TODO(iopi): think to remap methods to hooks
    if self.callObject == parsing.Parser.read_range:
        return "{}'{}'..'{}'".format(
            '\t' * level, self.params[0], self.params[1])
    elif self.callObject == parsing.Parser.read_char:
        return "{}'{}'".format('\t' * level, self.params[0])
    elif self.callObject == parsing.Parser.read_text:
        return "{}\"{}\"".format('\t' * level, self.params[0])
    elif self.callObject == parsing.Parser.read_integer:
        return "{}#num".format('\t' * level)
    elif self.callObject == parsing.Parser.read_identifier:
        return "{}#id".format('\t' * level)
    else:
        res = "{}#call: {} (".format('\t' * level, self.callObject.__name__)
        res += ", ".join(["{}".format(repr(param)) for param in self.params])
        res += ")"
        return res


@meta.add_method(parsing.Scope)
def to_dsl(self, level=0):
    res = "\n{}[{}\n".format('\t' * (level + 1), self.begin.to_dsl(0))
    res += self.pt.to_dsl(level + 1)
    res += "\n{}]{}\n".format('\t' * (level + 1), self.end.to_dsl(0))
    return res


@meta.add_method(parsing.Directive)
def to_dsl(self, level=0):
    # TODO: process parameter
    res = "\n{}@{}{}".format('\t' * (level + 1),
                             self.directive.ns_name,
                             _dumpParam(self))
    res += self.pt.to_dsl(level + 1)
    return res


@meta.add_method(parsing.Capture)
def to_dsl(self, level=0):
    res = "\n{}[\n".format('\t' * level)
    res += self.pt.to_dsl(level + 1)
    res += "\n{}] : {}\n".format('\t' * level, self.tagname)
    return res


@meta.add_method(parsing.Seq)
def to_dsl(self, level=0):
    return ' '.join([
        pt.to_dsl(level + 1) for pt in self.ptlist])


@meta.add_method(parsing.Alt)
def to_dsl(self, level=0):
    indent = '\t' * level
    res = "\n{}  {}".format(indent, self.ptlist[0].to_dsl(0))
    if len(self.ptlist) > 1:
        res += "\n{}| ".format(indent)
    res += "\n{}| ".format('\t' * level).join([
        pt.to_dsl(0) for pt in self.ptlist[1:]])
    return res


@meta.add_method(parsing.Neg)
def to_dsl(self, level=0):
    res = ("\n{}![\n".format('\t' * level))
    res += self.pt.to_dsl(level + 1)
    res += ("\n{}]\n".format('\t' * level))
    return res


@meta.add_method(parsing.Complement)
def to_dsl(self, level=0):
    res = ("\n{}~[\n".format('\t' * level))
    res += self.pt.to_dsl(level + 1)
    res += ("\n{}]\n".format('\t' * level))
    return res


@meta.add_method(parsing.Until)
def to_dsl(self, level=0):
    res = ("\n{}->[\n".format('\t' * level))
    res += self.pt.to_dsl(level + 1)
    res += ("\n{}]\n".format('\t' * level))
    return res


@meta.add_method(parsing.LookAhead)
def to_dsl(self, level=0):
    res = ("\n{}!![\n".format('\t' * level))
    res += self.pt.to_dsl(level + 1)
    res += ("\n{}]\n".format('\t' * level))
    return res


@meta.add_method(parsing.RepOptional)
def to_dsl(self, level=0):
    res = ("\n{}[\n".format('\t' * level))
    res += self.pt.to_dsl(level + 1)
    res += ("\n{}]?\n".format('\t' * level))
    return res


@meta.add_method(parsing.Rep0N)
def to_dsl(self, level=0):
    res = ("\n{}[\n".format('\t' * level))
    res += self.pt.to_dsl(level + 1)
    res += ("\n{}]*\n".format('\t' * level))
    return res


@meta.add_method(parsing.Rep1N)
def to_dsl(self, level=0):
    res = ("\n{}[\n".format('\t' * level))
    res += self.pt.to_dsl(level + 1)
    res += ("\n{}]+\n".format('\t' * level))
    return res



# error handling
import os
import atexit
import tempfile
import inspect
import re
import weakref
from collections import *
from pyrser import meta


Severity = meta.enum('INFO', 'WARNING', 'ERROR')


class LocationInfo:
    def __init__(self, filepath: str, line: int, col: int, size: int=1):
        self.filepath = filepath
        self.line = line
        self.col = col
        self.size = size

    @staticmethod
    def from_stream(stream: 'Stream', is_error=False) -> 'LocationInfo':
        if stream._name is None and is_error is True:
            (fh, stream._name) = tempfile.mkstemp()
            tmpf = os.fdopen(fh, 'w')
            tmpf.write(stream._content)
            tmpf.close()
            atexit.register(os.remove, stream._name)
        loc = LocationInfo(
            stream._name,
            stream._cursor.lineno,
            stream._cursor.col_offset
        )
        return loc

    @staticmethod
    def from_maxstream(stream: 'Stream', is_error=False) -> 'LocationInfo':
        if stream._name is None:
            (fh, stream._name) = tempfile.mkstemp()
            tmpf = os.fdopen(fh, 'w')
            tmpf.write(stream._content)
            tmpf.close()
            atexit.register(os.remove, stream._name)
        loc = LocationInfo(
            stream._name,
            stream._cursor._maxline,
            stream._cursor._maxcol
        )
        return loc

    @staticmethod
    def from_here(pos=1):
        f = inspect.currentframe()
        fcaller = inspect.getouterframes(f)[pos]
        rstr = r'(\s+).'
        cl = re.compile(rstr)
        call = fcaller[4][0]
        m = cl.match(call)
        current_file = os.path.abspath(fcaller[1])
        li = LocationInfo(current_file, fcaller[2], len(m.group(1)) + 1)
        return li

    def get_content(self) -> str:
        f = open(self.filepath, 'r')
        lines = list(f)
        f.close()
        txtline = lines[self.line - 1]
        indent = ' ' * (self.col - 1)
        if self.size != 1:
            indent += '~' * (self.size)
        else:
            indent += '^'
        txt = "from {f} at line:{l} col:{c} :\n{content}{i}".format(
            f=self.filepath,
            content=txtline,
            l=self.line,
            c=self.col,
            i=indent
        )
        return txt


class Notification:
    """
    Just One notification
    """
    def __init__(self, severity: Severity, msg: str,
                 location: LocationInfo=None, details: str=None):
        self.severity = severity
        self.location = location
        self.msg = msg
        self.details = details

    def get_content(self, with_locinfos=False, with_details=False) -> str:
        sevtxt = ""
        txt = "{s} : {msg}\n".format(
            s=Severity.rmap[self.severity].lower(),
            msg=self.msg
        )
        if with_locinfos and self.location is not None:
            txt += self.location.get_content()
        if with_details and self.details is not None:
            txt += self.details
        return txt


class Diagnostic(Exception):
    """
    The diagnostic object is use to handle easily
    all errors/warnings/infos in a compiler that you could
    encounter. Error while parsing, Error while type checking etc...
    You could use different severity for your notification.
    """
    def __init__(self):
        self.logs = []

    def __bool__(self):
        return self.have_errors

    def notify(self, severity: Severity, msg: str,
               location: LocationInfo=None, details: str=None) -> int:
        nfy = Notification(severity, msg, location, details)
        self.logs.append(nfy)
        return len(self.logs) - 1

    def add(self, n: Notification) -> int:
        if not isinstance(n, Notification):
            raise TypeError("Must be a notification")
        self.logs.append(n)
        return len(self.logs) - 1

    def get_content(self, with_locinfos=False, with_details=False) -> str:
        ls = []
        for v in self.logs:
            ls.append(v.get_content(with_locinfos, with_details))
        txt = ('=' * 79) + '\n'
        txt += ('\n' + ('-' * 79) + '\n').join(ls)
        txt += '\n' + ('-' * 79)
        return txt

    def get_infos(self) -> {Severity, int}:
        infos = dict()
        for s in Severity.map.values():
            infos[s] = 0
        for v in self.logs:
            s = v.severity
            infos[s] += 1
        return infos

    @property
    def have_errors(self) -> bool:
        for v in self.logs:
            if v.severity == Severity.ERROR:
                return True



import inspect
import types
from pyrser import meta, error
from pyrser.parsing.base import BasicParser
from pyrser.parsing.node import Node
from pyrser.parsing.stream import Tag


_decorators = []


class Functor:
    """Dummy Base class for all parse tree classes.

    common property:
        pt if contain a Functor
        ptlist if contain a list of Functor
    """

    def do_call(self, parser: BasicParser) -> Node:
        pass

    def __call__(self, parser: BasicParser) -> Node:
        global _decorators
        # call the begin methods in order
        for i in range(0, len(_decorators)):
            _decorators[i].begin(parser, self)
        # forward the call to the functor
        res = self.do_call(parser)
        # call the end methods in reverse order
        for i in range(len(_decorators) - 1, -1, -1):
            _decorators[i].end(res, parser, self)
        return res


class SkipIgnore(Functor):
    """Call Ignore Convention primitive functor."""

    def __init__(self, convention: str=""):
        """TODO: Could be better to implement Directive thru functors???"""
        self.convention = convention

    def do_call(self, parser: BasicParser) -> bool:
        #if len(parser._ignores) > 0:
        #    parser._ignores[-1](parser)
        parser.skip_ignore()
        return True


class PeekChar(Functor):
    """ !!'A' bnf primitive functor. """

    def __init__(self, c: str):
        self.char = c

    def do_call(self, parser: BasicParser) -> bool:
        return parser.peek_char(self.char)


class PeekText(Functor):
    """ !!"TXT" bnf primitive functor. """

    def __init__(self, c: str):
        self.char = c

    def do_call(self, parser: BasicParser) -> bool:
        return parser.peek_text(self.char)


class Text(Functor):
    """ "TXT" bnf primitive functor. """

    def __init__(self, txt: str):
        self.text = txt

    def do_call(self, parser: BasicParser) -> bool:
        return parser.read_text(self.text)


class Char(Functor):
    """ 'A' bnf primitive functor. """

    def __init__(self, c: str):
        self.char = c

    def do_call(self, parser: BasicParser) -> bool:
        return parser.read_char(self.char)


class Range(Functor):
    """ 'A'..'Z' bnf primitive functor. """

    def __init__(self, begin: str, end: str):
        self.begin = begin
        self.end = end

    def do_call(self, parser: BasicParser) -> bool:
        return parser.read_range(self.begin, self.end)


class UntilChar(Functor):
    """ ->'A' bnf primitive functor. """

    def __init__(self, c: str):
        self.char = c

    def do_call(self, parser: BasicParser) -> bool:
        return parser.read_until(self.char)


class Seq(Functor):
    """A B C bnf primitive as a functor."""

    def __init__(self, *ptlist: Functor):
        Functor.__init__(self)
        if len(ptlist) == 0:
            raise TypeError("Expected Functor")
        self.ptlist = []
        for it in ptlist:
            if not isinstance(it, SkipIgnore):
                self.ptlist.append(it)
                self.ptlist.append(SkipIgnore())
        if not isinstance(self.ptlist[0], SkipIgnore):
            self.ptlist.insert(0, SkipIgnore())

    def __getitem__(self, idx) -> Functor:
        """Hide SkipIgnore object from outside"""
        if idx >= 0:
            idx = (idx * 2) + 1
        else:
            idx = len(self.ptlist) - ((idx + 1) * 2) - 2
        return self.ptlist[idx]

    def do_call(self, parser: BasicParser) -> bool:
        parser._stream.save_context()
        for pt in self.ptlist:
            if not pt(parser):
                return parser._stream.restore_context()
        return parser._stream.validate_context()


class Scope(Functor):
    """functor to wrap SCOPE/rule directive or just []."""

    def __init__(self, begin: Seq, end: Seq, pt: Seq):
        Functor.__init__(self)
        self.begin = begin
        self.end = end
        self.pt = pt

    def do_call(self, parser: BasicParser) -> Node:
        if not self.begin(parser):
            return False
        res = self.pt(parser)
        if not self.end(parser):
            return False
        return res


class LookAhead(Functor):
    """!!A bnf primitive as a functor."""
    def __init__(self, pt: Functor):
        Functor.__init__(self)
        self.pt = pt
        if isinstance(self.pt, Seq):
            if isinstance(self.pt.ptlist[0], SkipIgnore):
                self.pt.ptlist.pop(0)
            if isinstance(self.pt.ptlist[-1], SkipIgnore):
                self.pt.ptlist.pop()

    def do_call(self, parser: BasicParser) -> bool:
        parser._stream.save_context()
        res = self.pt(parser)
        parser._stream.restore_context()
        return res


class Neg(Functor):
    """!A bnf primitive as a functor."""

    def __init__(self, pt: Functor):
        Functor.__init__(self)
        self.pt = pt
        if isinstance(self.pt, Seq):
            if isinstance(self.pt.ptlist[0], SkipIgnore):
                self.pt.ptlist.pop(0)
            if isinstance(self.pt.ptlist[-1], SkipIgnore):
                self.pt.ptlist.pop()

    def do_call(self, parser: BasicParser):
        parser._stream.save_context()
        if self.pt(parser):
            res = parser._stream.restore_context()
            return res
        return parser._stream.validate_context()


class Complement(Functor):
    """~A bnf primitive as a functor."""

    def __init__(self, pt: Functor):
        Functor.__init__(self)
        self.pt = pt
        if isinstance(self.pt, Seq):
            if isinstance(self.pt.ptlist[0], SkipIgnore):
                self.pt.ptlist.pop(0)
            if isinstance(self.pt.ptlist[-1], SkipIgnore):
                self.pt.ptlist.pop()

    def do_call(self, parser: BasicParser) -> bool:
        if parser.read_eof():
            return False
        parser._stream.save_context()
        res = self.pt(parser)
        if not res:
            parser._stream.incpos()
            return parser._stream.validate_context()
        parser._stream.restore_context()
        parser.undo_last_ignore()
        return False


class Until(Functor):
    """->A bnf primitive as a functor."""

    def __init__(self, pt: Functor):
        Functor.__init__(self)
        self.pt = pt
        if isinstance(self.pt, Seq):
            if isinstance(self.pt.ptlist[0], SkipIgnore):
                self.pt.ptlist.pop(0)
            if isinstance(self.pt.ptlist[-1], SkipIgnore):
                self.pt.ptlist.pop()

    def do_call(self, parser: BasicParser) -> bool:
        parser._stream.save_context()
        while not parser.read_eof():
            res = self.pt(parser)
            if res:
                return parser._stream.validate_context()
            parser._stream.incpos()
        parser._stream.restore_context()
        parser.undo_last_ignore()
        return False


class Call(Functor):
    """Functor wrapping a BasicParser method call in a BNF clause."""

    def __init__(self, callObject, *params):
        Functor.__init__(self)
        self.callObject = callObject
        self.params = params

    def do_call(self, parser: BasicParser) -> Node:
        return self.callObject(parser, *self.params)


class CallTrue(Call):
    """Functor to wrap arbitrary callable object in BNF clause."""

    def do_call(self, parser: BasicParser) -> Node:
        self.callObject(*self.params)
        return True


class Capture(Functor):
    """Functor to handle capture nodes."""

    def __init__(self, tagname: str, pt: Functor):
        Functor.__init__(self)
        if not isinstance(tagname, str) or len(tagname) == 0:
            raise TypeError("Illegal tagname for capture")
        self.tagname = tagname
        self.pt = pt
        if isinstance(self.pt, Seq):
            if isinstance(self.pt.ptlist[-1], SkipIgnore):
                self.pt.ptlist.pop()

    def do_call(self, parser: BasicParser) -> Node:
        if parser.begin_tag(self.tagname):
            # subcontext
            parser.push_rule_nodes()
            res = self.pt(parser)
            parser.pop_rule_nodes()
            if res and parser.end_tag(self.tagname):
                tag = parser.get_tag(self.tagname)
                # no bindings, wrap it in a Node instance
                if type(res) is bool:
                    res = Node()
                # update node cache
                parser.tag_node(self.tagname, res)
                parser.rule_nodes[self.tagname] = res
                # forward nodes
                return res
        return False


class DeclNode(Functor):
    """Functor to handle node declaration with __scope__:N."""

    def __init__(self, tagname: str):
        Functor.__init__(self)
        if not isinstance(tagname, str) or len(tagname) == 0:
            raise TypeError("Illegal tagname for capture")
        self.tagname = tagname

    def do_call(self, parser: BasicParser) -> Node:
        parser.rule_nodes[self.tagname] = Node()
        return True


class Bind(Functor):
    """Functor to handle the binding of a resulting nodes
    to an existing name.
    """

    def __init__(self, tagname: str, pt: Functor):
        Functor.__init__(self)
        if not isinstance(tagname, str) or len(tagname) == 0:
            raise TypeError("Illegal tagname for capture")
        self.tagname = tagname
        self.pt = pt

    def do_call(self, parser: BasicParser) -> Node:
        res = self.pt(parser)
        if res:
            parser.bind(self.tagname, res)
            return res
        return False


class Alt(Functor):
    """A | B bnf primitive as a functor."""

    def __init__(self, *ptlist: Seq):
        Functor.__init__(self)
        self.ptlist = ptlist

    def __getitem__(self, idx) -> Functor:
        return self.ptlist[idx]

    def do_call(self, parser: BasicParser) -> Node:
        # save result of current rule
        parser.push_rule_nodes()
        for pt in self.ptlist:
            parser._stream.save_context()
            parser.push_rule_nodes()
            res = pt(parser)
            if res:
                parser.pop_rule_nodes()
                parser.pop_rule_nodes()
                parser._stream.validate_context()
                return res
            parser.pop_rule_nodes()
            parser._stream.restore_context()
        parser.pop_rule_nodes()
        return False


class RepOptional(Functor):
    """[]? bnf primitive as a functor."""
    def __init__(self, pt: Seq):
        Functor.__init__(self)
        self.pt = pt
        if isinstance(self.pt, Directive):
            self.pt = Seq(self.pt)

    def do_call(self, parser: BasicParser) -> bool:
        res = self.pt(parser)
        if res:
            return res
        return True


class Rep0N(Functor):
    """[]* bnf primitive as a functor."""

    def __init__(self, pt: Seq):
        Functor.__init__(self)
        self.pt = pt
        if isinstance(self.pt, Directive):
            self.pt = Seq(self.pt)

    def do_call(self, parser: BasicParser) -> bool:
        parser.push_rule_nodes()
        while self.pt(parser):
            pass
        parser.pop_rule_nodes()
        return True


class Rep1N(Functor):
    """[]+ bnf primitive as a functor."""

    def __init__(self, pt: Seq):
        Functor.__init__(self)
        self.pt = pt
        if isinstance(self.pt, Directive):
            self.pt = Seq(self.pt)

    def do_call(self, parser: BasicParser) -> bool:
        parser._stream.save_context()
        parser.push_rule_nodes()
        if self.pt(parser):
            while self.pt(parser):
                pass
            parser.pop_rule_nodes()
            return parser._stream.validate_context()
        parser.pop_rule_nodes()
        return parser._stream.restore_context()


class Error(Functor):
    """Raise an error."""

    def __init__(self, msg: str, **kwargs):
        self.msg = msg
        self.kw = kwargs

    def do_call(self, parser: BasicParser) -> bool:
        parser.diagnostic.notify(
            error.Severity.ERROR,
            self.msg,
            error.LocationInfo.from_stream(parser._stream, is_error=True)
        )
        raise parser.diagnostic


class Rule(Functor):
    """Call a rule by its name."""

    def __init__(self, name: str):
        Functor.__init__(self)
        self.name = name

    def do_call(self, parser: BasicParser) -> Node:
        parser.push_rule_nodes()
        res = parser.eval_rule(self.name)
        parser.pop_rule_nodes()
        return res


class Hook(Functor):
    """Call a hook by his name."""

    def __init__(self, name: str, param: [(object, type)]):
        Functor.__init__(self)
        self.name = name
        # compose the list of value param, check type
        for v, t in param:
            if type(t) is not type:
                raise TypeError("Must be pair of value and type (i.e: int, "
                                "str, Node)")
        self.param = param

    def do_call(self, parser: BasicParser) -> bool:
        valueparam = []
        for v, t in self.param:
            if t is Node:
                if v not in parser.rule_nodes:
                    parser.diagnostic.notify(
                        error.Severity.ERROR,
                        "Unknown capture variable : %s" % v,
                        error.LocationInfo.from_stream(parser._stream, is_error=True)
                    )
                    raise parser.diagnostic
                valueparam.append(parser.rule_nodes[v])
            elif type(v) is t:
                valueparam.append(v)
            else:
                raise TypeError("Type mismatch expected {} got {}".format(
                    t, type(v)))
        return parser.eval_hook(self.name, valueparam)


class MetaDirectiveWrapper(type):
    """metaclass of all DirectiveWrapper subclasses.
    ensure that begin and end exists in subclasses as method"""
    def __new__(metacls, name, bases, namespace):
        cls = type.__new__(metacls, name, bases, namespace)
        if 'begin' not in namespace:
            raise TypeError(
                "DirectiveWrapper %s must have a begin method" % name)
        if not(isinstance(namespace['begin'], types.FunctionType)):
            raise TypeError(
                "'begin' not a function class in DirectiveWrapper %s" % name)
        if 'end' not in namespace:
            raise TypeError(
                "DirectiveWrapper %s subclasse must have a end method" % name)
        if not(isinstance(namespace['end'], types.FunctionType)):
            raise TypeError(
                "'end' not a function class in DirectiveWrapper %s" % name)
        return cls


class DirectiveWrapper(Functor, metaclass=MetaDirectiveWrapper):
    """functor to wrap begin/end directive"""

    def __init__(self, ):
        Functor.__init__(self)

    def checkParam(self, params: list):
        if (not hasattr(self.__class__, 'begin') or
                not hasattr(self.__class__, 'end')):
            return False
        sbegin = inspect.signature(self.begin)
        send = inspect.signature(self.end)

        idx = 0
        for param in list(sbegin.parameters.values())[1:]:
            if idx >= len(params) and param.default is inspect.Parameter.empty:
                raise RuntimeError("{}: No parameter given to begin"
                                   " method for argument {}, expected {}".
                                   format(
                                       self.__class__.__name__,
                                       idx, param.annotation))
            elif (idx < len(params)
                  and not isinstance(params[idx], param.annotation)):
                raise TypeError(
                    "{}: Wrong parameter in begin method parameter {} "
                    "expected {} got {}".format(
                        self.__class__.__name__,
                        idx, type(params[idx]),
                        param.annotation))
            idx += 1

        idx = 0
        for param in list(send.parameters.values())[1:]:
            if idx >= len(params) and param.default is inspect.Parameter.empty:
                raise RuntimeError("{}: No parameter given to end"
                                   " method for argument {}, expected {}".
                                   format(
                                       self.__class__.__name__,
                                       idx, param.annotation))
            elif (idx < len(params)
                  and not isinstance(params[idx], param.annotation)):
                raise TypeError(
                    "{}: Wrong parameter in end method parameter {} "
                    "expected {} got {}".format(
                        self.__class__.__name__,
                        idx, type(params[idx]),
                        param.annotation))
            idx += 1

        return True

    def begin(self):
        pass

    def end(self):
        pass


class Directive(Functor):
    """functor to wrap directive HOOKS"""

    def __init__(self, directive: DirectiveWrapper, param: [(object, type)],
                 pt: Functor):
        Functor.__init__(self)
        self.directive = directive
        self.pt = pt
        # compose the list of value param, check type
        for v, t in param:
            if type(t) is not type:
                raise TypeError(
                    "Must be pair of value and type (i.e: int, str, Node)")
        self.param = param

    def do_call(self, parser: BasicParser) -> Node:
        valueparam = []
        for v, t in self.param:
            if t is Node:
                valueparam.append(parser.rule_nodes[v])
            elif type(v) is t:
                valueparam.append(v)
            else:
                raise TypeError(
                    "Type mismatch expected {} got {}".format(t, type(v)))
        if not self.directive.checkParam(valueparam):
            return False
        if not self.directive.begin(parser, *valueparam):
            return False
        res = self.pt(parser)
        if not self.directive.end(parser, *valueparam):
            return False
        return res


class MetaDecoratorWrapper(type):
    """metaclass of all DecoratorWrapper subclasses.
    ensure that begin and end exists in subclasses as method"""
    def __new__(metacls, name, bases, namespace):
        cls = type.__new__(metacls, name, bases, namespace)
        if 'begin' not in namespace:
            raise TypeError(
                "DirectiveWrapper %s must have a begin method" % name)
        if not(isinstance(namespace['begin'], types.FunctionType)):
            raise TypeError(
                "'begin' not a function class in DirectiveWrapper %s" % name)
        if 'end' not in namespace:
            raise TypeError(
                "DirectiveWrapper %s subclasse must have a end method" % name)
        if not(isinstance(namespace['end'], types.FunctionType)):
            raise TypeError(
                "'end' not a function class in DirectiveWrapper %s" % name)
        return cls


class DecoratorWrapper(Functor, metaclass=MetaDecoratorWrapper):

    def begin(self):
        pass

    def end(self):
        pass


class Decorator(Functor):
    def __init__(self, decoratorClass: type, param: [(object, type)],
                 pt: Functor):
        self.decorator_class = decoratorClass
        self.pt = pt
        # compose the list of value param, check type
        for v, t in param:
            if type(t) is not type:
                raise TypeError(
                    "Must be pair of value and type (i.e: int, str, Node)")
        self.param = param

    def checkParam(self, the_class: type, params: list) -> bool:
        sinit = inspect.signature(the_class.__init__)

        idx = 0
        for param in list(sinit.parameters.values())[1:]:
            if idx >= len(params) and param.default is inspect.Parameter.empty:
                raise RuntimeError("{}: No parameter given to begin"
                                   " method for argument {}, expected {}".
                                   format(
                                       the_class.__name__,
                                       idx, param.annotation))
            elif (idx < len(params)
                  and not isinstance(params[idx], param.annotation)):
                raise TypeError(
                    "{}: Wrong parameter in begin method parameter {} "
                    "expected {} got {}".format(
                        the_class.__name__,
                        idx, type(params[idx]),
                        param.annotation))
            idx += 1

        return True

    def do_call(self, parser: BasicParser) -> Node:
        """
            The Decorator call is the one that actually pushes/pops
            the decorator in the active decorators list (parsing._decorators)
        """
        valueparam = []
        for v, t in self.param:
            if t is Node:
                valueparam.append(parser.rule_nodes[v])
            elif type(v) is t:
                valueparam.append(v)
            else:
                raise TypeError(
                    "Type mismatch expected {} got {}".format(t, type(v)))

        if not self.checkParam(self.decorator_class, valueparam):
            return False

        decorator = self.decorator_class(*valueparam)

        global _decorators
        _decorators.append(decorator)
        res = self.pt(parser)
        _decorators.pop()

        return res



# around a signature store type resolution for monomorphic or polymorphic call
import inspect
from pyrser import fmt
from pyrser.type_system.type_name import *
from pyrser.type_system.signature import *
from pyrser.type_system.fun import *
from pyrser.type_system.symbol import *


class EvalCtx:
    """
    Store environment for mono/poly call.
    """

    def __init__(self, sig: Signature):
        self._sig = sig
        self.resolution = dict()
        self._translate_to = None
        self._variadic_types = None
        self.resolve()

    def get_parent(self):
        return self._sig.get_parent()

    def internal_name(self):
        return self._sig.internal_name()

    def show_name(self):
        return self._sig.show_name()

    @property
    def name(self):
        if self._sig is not None:
            return self._sig.name
        return None

    @property
    def tret(self):
        if self._sig is not None and hasattr(self._sig, 'tret'):
            return self._sig.tret
        return None

    @property
    def tparams(self):
        if self._sig is not None and hasattr(self._sig, 'tparams'):
            return self._sig.tparams
        return None

    @property
    def variadic(self):
        if self._sig is not None and hasattr(self._sig, 'variadic'):
            return self._sig.variadic
        return None

    @property
    def is_polymorphic(self) -> bool:
        if self._sig is not None and hasattr(self._sig, 'is_polymorphic'):
            return self._sig.is_polymorphic
        return False

    def from_sig(sig: Signature) -> object:
        if isinstance(sig, EvalCtx):
            return sig
        else:
            return EvalCtx(sig)

    def set_parent(self, parent) -> object:
        """
        Forward to intern Signature
        """
        return self._sig.set_parent(parent)

    def __len__(self) -> int:
        return len(self._sig)

    def values(self) -> list:
        """
        Make EvalCtx iterable (to update a scope with it)
        """
        return [self._sig]

    def get_compute_sig(self) -> Signature:
        """
        Compute a signature Using resolution!!!

        TODO: discuss of relevance of a final generation for a signature
        """
        tret = []
        tparams = []
        for t in self.tret.components:
            if t in self.resolution and self.resolution[t] is not None:
                tret.append(self.resolution[t]().show_name())
            else:
                tret.append(t)
        if hasattr(self, 'tparams'):
            for p in self.tparams:
                tp = []
                for t in p.components:
                    if t in self.resolution and self.resolution[t] is not None:
                        tp.append(self.resolution[t]().show_name())
                    else:
                        tp.append(t)
                tparams.append(" ".join(tp))
            if self.variadic:
                if self._variadic_types is None:
                    raise ValueError("Can't compute the sig with unresolved variadic argument")
                for p in self._variadic_types:
                    tp = []
                    for t in p.components:
                        if t in self.resolution and self.resolution[t] is not None:
                            tp.append(self.resolution[t]().show_name())
                        else:
                            tp.append(t)
                    tparams.append(" ".join(tp))
        ret = Fun(self.name, " ".join(tret), tparams)
        # transform as-is into our internal Signature (Val, Var, whatever)
        ret.__class__ = self._sig.__class__
        return ret

    def set_parent(self, parent) -> object:
        """
        When we add a parent (from Symbol), don't forget to resolve.
        """
        ret = self
        if parent is not None:
            ret = self._sig.set_parent(parent)
            self.resolve()
        elif not hasattr(self, 'parent'):
            # only if parent didn't exist yet
            self.parent = None
        return ret

    def use_translator(self, translator):
        """
        Attach a translator to an EvalCtx to convert output of the function
        into another type
        """
        self._translate_to = translator
        self.resolve()

    def use_variadic_types(self, list_type: [TypeName]):
        """
        Attach a list of types for extra variadic argument of a call
        """
        self._variadic_types = list_type
        self.resolve()

    def resolve(self):
        """
        Process the signature and find definition for type.
        """
        # collect types for resolution
        t2resolv = []
        if hasattr(self._sig, 'tret'):
            t2resolv.append(self._sig.tret)
        if hasattr(self._sig, 'tparams') and self._sig.tparams is not None:
            for p in self._sig.tparams:
                t2resolv.append(p)
        if self._translate_to is not None:
            t2resolv.append(self._translate_to.target)
        if self._variadic_types is not None:
            for t in self._variadic_types:
                t2resolv.append(t)
        for t in t2resolv:
            for c in t.components:
                if c not in self.resolution or self.resolution[c] is None:
                    # try to find what is c
                    parent = self.get_parent()
                    if parent is not None:
                        sc = parent.get_by_symbol_name(c)
                        if len(sc) == 1:
                            sc = list(sc.values())[0]
                            # unwrap EvalCtx around Type
                            if isinstance(sc, EvalCtx):
                                sc = sc._sig
                            rtyp = weakref.ref(sc)
                            self.resolution[c] = rtyp
                            continue
                    # unresolved
                    self.resolution[c] = None

    def get_resolved_names(self, type_name: TypeName) -> list:
        """
        Use self.resolution to subsitute type_name.
        Allow to instanciate polymorphic type ?1, ?toto
        """
        if not isinstance(type_name, TypeName):
            raise Exception("Take a TypeName as parameter not a %s"
                            % type(type_name))
        rnames = []
        for name in type_name.components:
            if name not in self.resolution:
                raise Exception("Unknown type %s in a EvalCtx" % name)
            rname = self.resolution[name]
            if rname is not None:
                rname = rname().show_name()
            else:
                rname = name
            rnames.append(rname)
        return rnames

    def set_resolved_name(self, ref: dict, type_name2solve: TypeName,
                          type_name_ref: TypeName):
        """
        Warning!!! Need to rethink it when global poly type
        """
        if self.resolution[type_name2solve.value] is None:
            self.resolution[type_name2solve.value] = ref[type_name_ref.value]

    def to_fmt(self):
        """
        Return an Fmt representation for pretty-printing
        """
        qual = "evalctx"
        lseval = []
        block = fmt.block(":\n", "", fmt.tab(lseval))
        txt = fmt.sep(" ", [qual, block])
        lseval.append(self._sig.to_fmt())
        if len(self.resolution) > 0:
            lsb = []
            for k in sorted(self.resolution.keys()):
                s = self.resolution[k]
                if s is not None:
                    lsb.append(
                        fmt.end(
                            "\n",
                            ["'%s': %s (%s)" % (k, s, s().show_name())]
                        )
                    )
                else:
                    lsb.append(fmt.end("\n", ["'%s': Unresolved" % (k)]))
            if self._translate_to is not None:
                lsb.append("use translator:")
                lsb.append(self._translate_to.to_fmt())
            if self._variadic_types is not None:
                lsb.append("variadic types:\n")
                arity = self._sig.arity
                for t in self._variadic_types:
                    lsb.append("[%d] : %s\n" % (arity, t))
                    arity += 1
            lseval.append(fmt.block("\nresolution :\n", "", fmt.tab(lsb)))
        return txt

    def __str__(self):
        return str(self.to_fmt())



# tuple handling
from pyrser import fmt
from pyrser.type_system.signature import *


class Tuple(list):
    """
    A tuple is just a list of signature.
    """
    def __init__(self, sig: [Signature]=None):
        self._lsig = []
        if sig is not None:
            if isinstance(sig, Signature):
                self.add(sig)
            elif len(sig) > 0:
                for s in sig:
                    self.add(s)

    def add(self, sig: Signature):
        self._lsig.append(sig)

    def to_fmt(self) -> fmt.indentable:
        """
        Return an Fmt representation for pretty-printing
        """
        lsb = []
        if len(self._lsig) > 0:
            for s in self._lsig:
                lsb.append(s.to_fmt())
        block = fmt.block("(", ")", fmt.sep(', ', lsb))
        qual = "tuple"
        txt = fmt.sep("", [qual, block])
        return txt

    def __repr__(self) -> str:
        """
        Internal representation
        """
        return repr(self._lsig)

    def __str__(self) -> str:
        """
        Usefull representation
        """
        return str(self.to_fmt())



# symbol for type checking
import weakref


class Symbol:
    """
    Symbol are semantic name used in a language.
    Could be map to binary symbol but conceptually are more related
    to language symbol table.
    """

    def __init__(self, name, parent=None):
        self.name = name
        self.set_parent(parent)

    def __getstate__(self):
        """
        For pickle don't handle weakrefs...
        """
        state = self.__dict__.copy()
        del state['parent']
        return state

    def set_parent(self, parent) -> object:
        if parent is not None:
            self.parent = weakref.ref(parent)
        else:
            self.parent = None
        return self

    def get_parent(self) -> object:
        """
        Auto deref parent and return the instance.
        """
        if self.parent is not None:
            return self.parent()
        return None

    def get_scope_list(self) -> list:
        """
        Return the list of all contained scope from global to local
        """
        # by default only return scoped name
        lstparent = [self]
        p = self.get_parent()
        while p is not None:
            lstparent.append(p)
            p = p.get_parent()
        return lstparent

    def get_scope_names(self) -> list:
        """
        Return the list of all contained scope from global to local
        """
        # allow global scope to have an None string instance
        lscope = []
        for scope in reversed(self.get_scope_list()):
            if scope.name is not None:
                # handle fun/block scope decoration
                lscope.append(scope.name)
        return lscope

    # to overload for new language
    def show_name(self) -> str:
        """
        Return a convenient name for type checking
        """
        return ".".join(self.get_scope_names())

    # to overload for new language
    def internal_name(self) -> str:
        """
        Returns the namespace's internal_name.
        """
        unq = "_".join(self.get_scope_names())
        return unq

    def __str__(self) -> str:
        return self.show_name()

    def __repr__(self) -> str:
        return self.internal_name()



# This pass is for debug only
from pyrser import meta, fmt
from pyrser.parsing import node
import weakref

scalar = {'bool': bool, 'int': int, 'float': float, 'str': str}
composed = {'list': list, 'dict': dict, 'set': set}


@meta.add_method(node.Node)
def to_yml(self):
    pp = fmt.tab([])
    to_yml_item(self, pp.lsdata, "")
    return str(pp)


def yml_attr(k, v):
    return fmt.sep("=", [k, v])


def to_yml_item(item, pp, name):
    global scalar
    refcount = weakref.getweakrefcount(item)
    if refcount > 0:
        name += " &" + str(id(item))
    if type(item).__name__ in scalar:
        tag = fmt.end(
            '\n',
            fmt.sep(
                "",
                [
                    name,
                    " ",
                    yml_attr(
                        type(item).__name__,
                        repr(item)
                    )
                ]
            )
        )
        pp.append(tag)
        return
    if isinstance(item, weakref.ref):
        name += " *" + str(id(item()))
        tag = fmt.end('\n', fmt.sep("", [name, " ", repr(item)]))
        pp.append(tag)
        return
    if isinstance(item, bytes) or isinstance(item, bytearray):
        inner = fmt.tab([])
        tag = fmt.block(name + " " + str(yml_attr('type', 'bytes')) + ':\n',
                        '----' + name + '----\n', inner)
        inner.lsdata.append(fmt.sep(" ", []))
        bindata = inner.lsdata[-1].lsdata
        i = 0
        for b in item:
            bindata.append("%02X" % b)
            i += 1
            if i > 16:
                bindata.append("\n")
                i = 0
        bindata.append("\n")
        pp.append(tag)
        return
    if isinstance(item, object) and hasattr(item, '__dict__'):
        inner = fmt.tab([])
        tag = fmt.block(name + " " +
                        str(yml_attr('type', item.__class__.__name__))
                        + ':\n',
                        '', inner)
        for attr in sorted(vars(item)):
            to_yml_item(getattr(item, attr), inner.lsdata, attr)
        if len(vars(item)) == 0:
            inner.lsdata.append("\n")
        pp.append(tag)
        return
    if isinstance(item, list):
        inner = fmt.tab([])
        tag = fmt.block(
            name + " " + str(yml_attr('type', 'list')) + ':\n',
            '',
            inner
        )
        i = 0
        for subitem in item:
            idxname = str(fmt.sep(" ", ['[', i, ']']))
            to_yml_item(subitem, inner.lsdata, idxname)
            i += 1
        if len(item) == 0:
            inner.lsdata.append("\n")
        pp.append(tag)
        return
    if isinstance(item, tuple):
        inner = fmt.tab([])
        tag = fmt.block(name + " " + str(yml_attr('type', 'tuple')) + ':\n',
                        '', inner)
        i = 0
        for subitem in item:
            idxname = str(fmt.sep(" ", ["[", i, "]"]))
            to_yml_item(subitem, inner.lsdata, idxname)
            i += 1
        if len(item) == 0:
            inner.lsdata.append("\n")
        pp.append(tag)
        return
    if isinstance(item, dict):
        inner = fmt.tab([])
        tag = fmt.block(name + " " + str(yml_attr('type', 'dict')) + ':\n',
                        '', inner)
        for k in sorted(item.keys()):
            idxname = str(fmt.sep(" ", ["[", repr(k), "]"]))
            to_yml_item(item[k], inner.lsdata, idxname)
        if len(item.keys()) == 0:
            inner.lsdata.append("\n")
        pp.append(tag)
        return
    if isinstance(item, set):
        inner = fmt.tab([])
        tag = fmt.block(name + " " + str(yml_attr('type', 'set')) + ':\n',
                        '', inner)
        for subitem in sorted(item):
            inner.lsdata.append(fmt.sep(", "[repr(subitem)]))
        if len(item) == 0:
            inner.lsdata.append("\n")
        pp.append(tag)
        return
    if item is None:
        tag = fmt.end('\n', [name])
        pp.append(tag)
        return



# qualified (or not) type names


class TypeName(str):

    def __init__(self, value):
        super().__init__()
        self.value = value
        # split value into composed type and/or qualifiers
        self.components = value.split()

    def __str__(self) -> str:
        return self.value

    def __hash__(self) -> int:
        return self.value.__hash__()

    def __lt__(self, oth) -> bool:
        if isinstance(oth, TypeName):
            return self.value < oth.value
        return self.value < oth

    def __eq__(self, oth) -> bool:
        if isinstance(oth, TypeName):
            return self.value == oth.value
        return self.value == oth

    def get_subcomponents(self) -> list:
        return self.components[1:]

    @property
    def is_polymorphic(self) -> bool:
        """
        Check if one of composed type name is poly
        """
        for c in self.components:
            if c[0] == '?':
                return True
        return False



from pyrser import parsing
from pyrser import meta
from pyrser import error
from pyrser.directives import ignore


class EBNF(parsing.Parser):
    """
    Basic class for BNF DSL PARSING.

    A full parser for the BNF is provided by this class.
    We construct a tree to represents, thru functors, BNF semantics.
    """

    def get_rules(self) -> parsing.Node:
        """
        Parse the DSL and provide a dictionnaries of all resulting rules.
        Call by the MetaGrammar class.
        """
        # TODO:
        res = None
        try:
            res = self.eval_rule('bnf_dsl')
            if not res:
                # we fail to parse, but error is not set
                self.diagnostic.notify(
                    error.Severity.ERROR,
                    "Parse error in '%s' in EBNF bnf" % self._lastRule,
                    error.LocationInfo.from_maxstream(self._stream)
                )
                raise self.diagnostic
        except error.Diagnostic as d:
            d.notify(
                error.Severity.ERROR,
                "Parse error in '%s' in EBNF bnf" % self._lastRule
            )
            raise d
        return res

    def __init__(self, content='', sname=None):
        """
        Define the DSL parser.
        """
        super().__init__(content, sname)
        self.set_rules({
            #
            # bnf_dsl = [ @ignore("C/C++") bnf_stmts ]
            #
            'bnf_dsl': parsing.Seq(
                # tree is not already construct but Directive need it
                # forward it thru a lambda
                parsing.Directive(ignore.Ignore(),
                                  [("C/C++", str)],
                                  lambda parser:
                                  self.__class__._rules['bnf_stmts'](parser)),
            ),

            #
            # bnf_stmts = [ [rule : r #add_rules(_, r) ]+ Base.eof ]
            #
            'bnf_stmts': parsing.Seq(
                parsing.Rep1N(parsing.Seq(
                    parsing.Capture("r", parsing.Rule('rule')),
                    parsing.Hook('add_rules', [("_", parsing.Node),
                                               ("r", parsing.Node)])
                )),
                parsing.Rule('Base.eof')
            ),

            # TODO: add directive hooks / change ns_name by def_rule
            #
            # rule = [ ns_name : rn '=' '[' alternatives : alts
            #                             #add_rule(_, rn, alts) ']' ]
            #
            'rule': parsing.Seq(
                parsing.Capture("rn", parsing.Rule('ns_name')),
                parsing.Alt(
                    #parsing.Call(parsing.Parser.read_text, "="),
                    parsing.Char("="),
                    parsing.Error("Expected '='")),
                parsing.Alt(
                    #parsing.Call(parsing.Parser.read_text, "["),
                    parsing.Char("["),
                    parsing.Error("Expected '['")),
                parsing.Capture("alts", parsing.Rule('alternatives')),
                parsing.Hook('add_rule', [("_", parsing.Node),
                                          ("rn", parsing.Node),
                                          ("alts", parsing.Node)]),
                parsing.Alt(
                    #parsing.Call(parsing.Parser.read_char, ']'),
                    parsing.Char(']'),
                    parsing.Error("Expected ']'"))
            ),

            #
            # alternatives =
            # [
            #       sequences : alt #add_alt(_, alt)
            #       ['|' sequences : alt #add_alt(_, alt) ]*
            # ]
            #
            'alternatives': parsing.Seq(
                parsing.Capture('alt', parsing.Rule('sequences')),
                parsing.Hook('add_alt', [("_", parsing.Node),
                                         ("alt", parsing.Node)]),
                parsing.Rep0N(
                    parsing.Seq(
                        #parsing.Call(parsing.Parser.read_char, '|'),
                        parsing.Char('|'),
                        parsing.Capture('alt', parsing.Rule('sequences')),
                        parsing.Hook('add_alt',
                                     [("_", parsing.Node),
                                      ("alt", parsing.Node)])
                    )
                )
            ),

            #
            # sequences = [ [ sequence : cla #add_sequences(_, cla) ]+ ]
            #
            'sequences': parsing.Rep1N(
                parsing.Seq(
                    parsing.Capture('cla', parsing.Rule('sequence')),
                    parsing.Hook('add_sequences',
                                 [("_", parsing.Node),
                                  ("cla", parsing.Node)])
                )
            ),

            #
            # sequence = [
            #   [
            #     [ '~' | "!!" | '!' | "->" ]?: mod
            #     [ ns_name : rid #add_ruleclause_name(_, rid)
            #       | Base.string : txt #add_text(_, txt)
            #       | Base.char : begin ".." Base.char : end
            #         #add_range(_, begin, end)
            #       | Base.char : c #add_char(_, c)
            #       | '[' alternatives : subsequence ']'
            #         #add_subsequence(_, subsequence)
            #     ] #add_mod(_, mod)
            #     [ repeat : rpt #add_rpt(_, mod, rpt) ]?
            #   | hook : h #add_hook(_, h)
            #   | directive : d sequences : s #add_directive(_, d, s)
            #   ]
            #     [
            #       ":>" Base.id : bind #add_bind(_, bind)
            #       | ':' Base.id : cpt #add_capture(_, cpt)
            #     ]?
            # ]
            #
            'sequence':
            parsing.Seq(
                parsing.Alt(
                    parsing.Seq(
                        parsing.Capture(
                            'mod',
                            parsing.RepOptional(
                                parsing.Alt(
                                    #parsing.Call(
                                    #    parsing.Parser.read_char,
                                    #    '~'
                                    #),
                                    parsing.Char('~'),
                                    #parsing.Call(
                                    #    parsing.Parser.read_text,
                                    #    '!!'
                                    #),
                                    parsing.Text('!!'),
                                    #parsing.Call(
                                    #    parsing.Parser.read_char,
                                    #    '!'
                                    #),
                                    parsing.Char('!'),
                                    #parsing.Call(
                                    #    parsing.Parser.read_text,
                                    #    '->'
                                    #)
                                    parsing.Text('->')
                                )
                            )
                        ),
                        parsing.Alt(
                            parsing.Seq(
                                parsing.Capture(
                                    'rid',
                                    parsing.Rule('ns_name')
                                ),
                                parsing.Hook('add_ruleclause_name',
                                             [("_", parsing.Node),
                                              ("rid", parsing.Node)])
                            ),
                            parsing.Seq(
                                parsing.Capture('txt',
                                                parsing.Rule('Base.string')),
                                parsing.Hook('add_text',
                                             [("_", parsing.Node),
                                              ("txt", parsing.Node)])
                            ),
                            parsing.Seq(
                                parsing.Capture('begin',
                                                parsing.Rule('Base.char')),
                                #parsing.Call(parsing.Parser.read_text, ".."),
                                parsing.Text(".."),
                                parsing.Capture(
                                    'end',
                                    parsing.Rule('Base.char')
                                ),
                                parsing.Hook('add_range',
                                             [("_", parsing.Node),
                                              ("begin", parsing.Node),
                                              ("end", parsing.Node)])
                            ),
                            parsing.Seq(
                                parsing.Capture(
                                    'c',
                                    parsing.Rule('Base.char')
                                ),
                                parsing.Hook('add_char',
                                             [("_", parsing.Node),
                                              ("c", parsing.Node)])
                            ),
                            parsing.Seq(
                                #parsing.Call(parsing.Parser.read_char, "["),
                                parsing.Char('['),
                                parsing.Capture(
                                    'subsequence',
                                    parsing.Alt(
                                        parsing.Rule('alternatives'),
                                        parsing.Error("Expected sequences"))),
                                parsing.Alt(
                                    #parsing.Call(
                                    #    parsing.Parser.read_char,
                                    #    "]"
                                    #),
                                    parsing.Char(']'),
                                    parsing.Error("Expected ']'")),
                                parsing.Hook('add_subsequence',
                                             [("_", parsing.Node),
                                              ("subsequence", parsing.Node)]),
                            )
                        ),
                        parsing.Hook('add_mod', [("_", parsing.Node),
                                                 ("mod", parsing.Node)]),
                        parsing.RepOptional(
                            parsing.Seq(
                                parsing.Capture(
                                    'rpt',
                                    parsing.Rule('repeat')
                                ),
                                parsing.Hook('add_rpt',
                                             [("_", parsing.Node),
                                              ("mod", parsing.Node),
                                              ("rpt", parsing.Node)])
                            )
                        ),
                    ),
                    parsing.Seq(
                        parsing.Capture('h', parsing.Rule('hook')),
                        parsing.Hook('add_hook', [('_', parsing.Node),
                                                  ('h', parsing.Node)])
                    ),
                    parsing.Seq(
                        parsing.Capture('d', parsing.Rule('directive')),
                        parsing.Capture('s', parsing.Rule('sequences')),
                        parsing.Hook('add_directive', [('_', parsing.Node),
                                                       ('d', parsing.Node),
                                                       ('s', parsing.Node)])
                    )
                ),
                parsing.RepOptional(
                    parsing.Alt(
                        parsing.Seq(
                            #parsing.Call(parsing.Parser.read_text, ":>"),
                            parsing.Text(':>'),
                            parsing.Capture(
                                'bind',
                                parsing.Rule('Base.id')),
                            parsing.Hook('add_bind',
                                         [('_', parsing.Node),
                                          ('bind', parsing.Node)])
                        ),
                        parsing.Seq(
                            #parsing.Call(parsing.Parser.read_text, ":"),
                            parsing.Char(':'),
                            parsing.Capture(
                                'cpt',
                                parsing.Rule('Base.id')),
                            parsing.Hook('add_capture',
                                         [('_', parsing.Node),
                                          ('cpt', parsing.Node)])
                        )
                    )
                )
            ),

            # ns_name = [ [@ignore("null") [ Base.id ['.' Base.id]* ]]: rid ]
            #
            'ns_name': parsing.Capture(
                'rid',
                parsing.Scope(
                    parsing.Call(parsing.Parser.push_ignore,
                                 parsing.Parser.ignore_null),
                    parsing.Call(parsing.Parser.pop_ignore),
                    parsing.Seq(
                        parsing.Rule('Base.id'),
                        parsing.Rep0N(
                            parsing.Seq(
                                #parsing.Call(parsing.Parser.read_text, "."),
                                parsing.Char('.'),
                                parsing.Alt(
                                    parsing.Rule('Base.id'),
                                    parsing.Error(
                                        "Expected identifier after '.'"))
                            )
                        )
                    )
                )
            ),

            #
            # repeat = [ '?' #add_optional(_)
            #          | '*' #add_0N(_)
            #          | '+' #add_1N(_)
            # ]
            #
            'repeat': parsing.Alt(
                parsing.Seq(
                    #parsing.Call(parsing.Parser.read_char, '?'),
                    parsing.Char('?'),
                    parsing.Hook('add_optional', [("_", parsing.Node)])
                ),
                parsing.Seq(
                    #parsing.Call(parsing.Parser.read_char, '*'),
                    parsing.Char('*'),
                    parsing.Hook('add_0N', [("_", parsing.Node)])
                ),
                parsing.Seq(
                    #parsing.Call(parsing.Parser.read_char, '+'),
                    parsing.Char('+'),
                    parsing.Hook('add_1N', [("_", parsing.Node)])
                ),
            ),

            #
            # hook = [ '#' ns_name : n #hook_name(_, n)
            #          ['(' param : p #hook_param(_, p)
            #              [',' param : p #hook_param(_, p)]*
            #           ')']?
            # ]
            #
            'hook': parsing.Seq(
                #parsing.Call(parsing.Parser.read_char, '#'),
                parsing.Char('#'),
                parsing.Capture('n', parsing.Rule('ns_name')),
                parsing.Hook('hook_name',
                             [('_', parsing.Node), ('n', parsing.Node)]),
                parsing.RepOptional(
                    parsing.Seq(
                        #parsing.Call(parsing.Parser.read_char, '('),
                        parsing.Char('('),
                        parsing.Capture(
                            'p',
                            parsing.Alt(
                                parsing.Rule('param'),
                                parsing.Error("Expected parameter"))),
                        parsing.Hook('hook_param', [('_', parsing.Node),
                                                    ('p', parsing.Node)]),
                        parsing.Rep0N(
                            parsing.Seq(
                                #parsing.Call(parsing.Parser.read_char, ','),
                                parsing.Char(','),
                                parsing.Capture(
                                    'p',
                                    parsing.Alt(
                                        parsing.Rule('param'),
                                        parsing.Error("Expected parameter"))),
                                parsing.Hook('hook_param',
                                             [('_', parsing.Node),
                                              ('p', parsing.Node)]))
                        ),
                        parsing.Alt(
                            #parsing.Call(parsing.Parser.read_char, ')'),
                            parsing.Char(')'),
                            parsing.Error("Expected ')'"))
                    )
                ),
            ),

            #
            # directive = [ '@' ns_name : n #hook_name(_, n)
            #               ['(' param : p #hook_param(_, p)
            #                   [',' param : p #hook_param(_, p)]*
            #                ')']?
            # ]
            'directive': parsing.Seq(
                #parsing.Call(parsing.Parser.read_char, '@'),
                parsing.Char('@'),
                parsing.Capture('n', parsing.Rule('ns_name')),
                parsing.Hook('hook_name', [('_', parsing.Node),
                                           ('n', parsing.Node)]),
                parsing.RepOptional(
                    parsing.Seq(
                        #parsing.Call(parsing.Parser.read_char, '('),
                        parsing.Char('('),
                        parsing.Capture(
                            'p',
                            parsing.Alt(
                                parsing.Rule('param'),
                                parsing.Error("Expected parameter"))),
                        parsing.Hook('hook_param',
                                     [('_', parsing.Node),
                                      ('p', parsing.Node)]),
                        parsing.Rep0N(
                            parsing.Seq(
                                #parsing.Call(parsing.Parser.read_char, ','),
                                parsing.Char(','),
                                parsing.Capture(
                                    'p',
                                    parsing.Alt(
                                        parsing.Rule('param'),
                                        parsing.Error("Expected parameter"))),
                                parsing.Hook('hook_param',
                                             [('_', parsing.Node),
                                              ('p', parsing.Node)]),
                            )
                        ),
                        parsing.Alt(
                            #parsing.Call(parsing.Parser.read_char, ')'),
                            parsing.Char(')'),
                            parsing.Error("Expected ')'"))
                    )
                ),
            ),

            #
            # param = [ Base.num :n #param_num(_, n)
            #         | Base.string : s #param_str(_, s)
            #         | Base.char : c #param_char(_, c)
            #         | ns_name : i #param_id(_, i)
            # ]
            #
            'param': parsing.Alt(
                parsing.Seq(
                    parsing.Capture('n', parsing.Rule('Base.num')),
                    parsing.Hook('param_num', [('_', parsing.Node),
                                               ('n', parsing.Node)])
                ),
                parsing.Seq(
                    parsing.Capture('s', parsing.Rule('Base.string')),
                    parsing.Hook('param_str', [('_', parsing.Node),
                                               ('s', parsing.Node)])
                ),
                parsing.Seq(
                    parsing.Capture('c', parsing.Rule('Base.char')),
                    parsing.Hook('param_char', [('_', parsing.Node),
                                                ('c', parsing.Node)])
                ),
                parsing.Seq(
                    parsing.Capture('i', parsing.Rule('ns_name')),
                    parsing.Hook('param_id', [('_', parsing.Node),
                                              ('i', parsing.Node)])
                ),
            ),
        })

# Hooks part
# ----------
# All these functions are automatically added to class EBNF


@meta.hook(EBNF, "EBNF.add_mod")
def add_mod(self, seq, mod):
    """Create a tree.{Complement, LookAhead, Neg, Until}"""
    modstr = self.value(mod)
    if modstr == '~':
        seq.parser_tree = parsing.Complement(seq.parser_tree)
    elif modstr == '!!':
        seq.parser_tree = parsing.LookAhead(seq.parser_tree)
    elif modstr == '!':
        seq.parser_tree = parsing.Neg(seq.parser_tree)
    elif modstr == '->':
        seq.parser_tree = parsing.Until(seq.parser_tree)
    return True


@meta.hook(EBNF, "EBNF.add_ruleclause_name")
def add_ruleclause_name(self, ns_name, rid) -> bool:
    """Create a tree.Rule"""
    ns_name.parser_tree = parsing.Rule(self.value(rid))
    return True


@meta.hook(EBNF, "EBNF.add_rules")
def add_rules(self, bnf, r) -> bool:
    """Attach a parser tree to the dict of rules"""
    bnf[r.rulename] = r.parser_tree
    return True


@meta.hook(EBNF, "EBNF.add_rule")
def add_rule(self, rule, rn, alts) -> bool:
    """Add the rule name"""
    rule.rulename = self.value(rn)
    rule.parser_tree = alts.parser_tree
    return True


@meta.hook(EBNF, "EBNF.add_sequences")
def add_sequences(self, sequences, cla) -> bool:
    """Create a tree.Seq"""
    if not hasattr(sequences, 'parser_tree'):
        # forward sublevel of sequence as is
        sequences.parser_tree = cla.parser_tree
    else:
        oldnode = sequences
        if isinstance(oldnode.parser_tree, parsing.Seq):
            oldpt = list(oldnode.parser_tree.ptlist)
        else:
            oldpt = [oldnode.parser_tree]
        oldpt.append(cla.parser_tree)
        sequences.parser_tree = parsing.Seq(*tuple(oldpt))
    return True


@meta.hook(EBNF, "EBNF.add_alt")
def add_alt(self, alternatives, alt) -> bool:
    """Create a tree.Alt"""
    if not hasattr(alternatives, 'parser_tree'):
        # forward sublevel of alt as is
        if hasattr(alt, 'parser_tree'):
            alternatives.parser_tree = alt.parser_tree
        else:
            alternatives.parser_tree = alt
    else:
        oldnode = alternatives
        if isinstance(oldnode.parser_tree, parsing.Alt):
            oldpt = list(oldnode.parser_tree.ptlist)
        else:
            oldpt = [oldnode.parser_tree]
        oldpt.append(alt.parser_tree)
        alternatives.parser_tree = parsing.Alt(*tuple(oldpt))
    return True


@meta.hook(EBNF, "EBNF.add_char")
def add_char(self, sequence, c):
    """Add a read_char primitive"""
    sequence.parser_tree = parsing.Char(self.value(c).strip("'"))
    return True


@meta.hook(EBNF, "EBNF.add_text")
def add_text(self, sequence, txt):
    """Add a read_text primitive"""
    sequence.parser_tree = parsing.Text(self.value(txt).strip('"'))
    return True


@meta.hook(EBNF, "EBNF.add_range")
def add_range(self, sequence, begin, end):
    """Add a read_range primitive"""
    sequence.parser_tree = parsing.Range(self.value(begin).strip("'"),
                                        self.value(end).strip("'"))
    return True


@meta.hook(EBNF, "EBNF.add_rpt")
def add_rpt(self, sequence, mod, pt):
    """Add a repeater to the previous sequence"""
    modstr = self.value(mod)
    if modstr == '!!':
        # cursor on the REPEATER
        self._stream.restore_context()
        # log the error
        self.diagnostic.notify(
            error.Severity.ERROR,
            "Cannot repeat a lookahead rule",
            error.LocationInfo.from_stream(self._stream, is_error=True)
        )
        raise self.diagnostic
    if modstr == '!':
        # cursor on the REPEATER
        self._stream.restore_context()
        # log the error
        self.diagnostic.notify(
            error.Severity.ERROR,
            "Cannot repeat a negated rule",
            error.LocationInfo.from_stream(self._stream, is_error=True)
        )
        raise self.diagnostic
    oldnode = sequence
    sequence.parser_tree = pt.functor(oldnode.parser_tree)
    return True


@meta.hook(EBNF, "EBNF.add_capture")
def add_capture(self, sequence, cpt):
    """Create a tree.Capture"""
    cpt_value = self.value(cpt)
    sequence.parser_tree = parsing.Capture(cpt_value, sequence.parser_tree)
    return True


@meta.hook(EBNF, "EBNF.add_bind")
def add_bind(self, sequence, cpt):
    """Create a tree.Bind"""
    cpt_value = self.value(cpt)
    sequence.parser_tree = parsing.Bind(cpt_value, sequence.parser_tree)
    return True


@meta.hook(EBNF, "EBNF.add_subsequence")
def add_subsequence(self, sequence, subsequence):
    """Add a subsequence into a sequence"""
    sequence.parser_tree = subsequence.parser_tree
    return True


@meta.hook(EBNF, "EBNF.add_optional")
def add_optional(self, repeat):
    """Create a tree.RepOptional"""
    repeat.functor = parsing.RepOptional
    return True


@meta.hook(EBNF, "EBNF.add_0N")
def add_0N(self, repeat):
    """Create a tree.Rep0N"""
    repeat.functor = parsing.Rep0N
    return True


@meta.hook(EBNF, "EBNF.add_1N")
def add_1N(self, repeat):
    """Create a tree.Rep1N"""
    repeat.functor = parsing.Rep1N
    return True


@meta.hook(EBNF, "EBNF.add_hook")
def add_hook(self, sequence, h):
    """Create a tree.Hook"""
    sequence.parser_tree = parsing.Hook(h.name, h.listparam)
    return True


@meta.hook(EBNF, "EBNF.param_num")
def param_num(self, param, n):
    """Parse a int in parameter list"""
    param.pair = (int(self.value(n)), int)
    return True


@meta.hook(EBNF, "EBNF.param_str")
def param_str(self, param, s):
    """Parse a str in parameter list"""
    param.pair = (self.value(s).strip('"'), str)
    return True


@meta.hook(EBNF, "EBNF.param_char")
def param_char(self, param, c):
    """Parse a char in parameter list"""
    param.pair = (self.value(c).strip("'"), str)
    return True


@meta.hook(EBNF, "EBNF.param_id")
def param_id(self, param, i):
    """Parse a node name in parameter list"""
    param.pair = (self.value(i), parsing.Node)
    return True


@meta.hook(EBNF, "EBNF.hook_name")
def hook_name(self, hook, n):
    """Parse a hook name"""
    hook.name = self.value(n)
    hook.listparam = []
    return True


@meta.hook(EBNF, "EBNF.hook_param")
def hook_param(self, hook, p):
    """Parse a hook parameter"""
    hook.listparam.append(p.pair)
    return True


@meta.hook(EBNF, "EBNF.add_directive")
def add_directive(self, sequence, d, s):
    """Add a directive in the sequence"""
    if d.name in meta._directives:
        the_class = meta._directives[d.name]
        sequence.parser_tree = parsing.Directive(the_class(), d.listparam,
                                                 s.parser_tree)
    elif d.name in meta._decorators:
        the_class = meta._decorators[d.name]
        sequence.parser_tree = parsing.Decorator(the_class, d.listparam,
                                                 s.parser_tree)
    else:
        raise TypeError("Unkown directive or decorator %s" % d.name)
    return True



class Node(dict):
    """Base class for node manipulation."""

    def __bool__(self):
        return True

    def __repr__(self):
        #TODO: use to_yml
        items = []
        if len(self) > 0:
            items.append(repr(self.items()))
        for k, v in vars(self).items():
            items.append("{} = {}".format(k, repr(v)))
        return "{}({})".format(self.__class__.__name__, ', '.join(items))

    def check(self, ndict: dict, info="") -> bool:
        """
        Debug method, help detect cycle and/or
        other incoherence in a tree of Node
        """
        def iscycle(thing, ndict: dict, info: str) -> bool:
            # check if not already here
            idthing = id(thing)
            ndict[info] = idthing
            if idthing not in ndict:
                # add myself
                ndict[idthing] = "%s:%s no cycle" % (type(thing), info)
                return False
            else:
                ndict[idthing] += "\n%s:%s cycle" % (type(thing), info)
                return True

        def recurs(thing, ndict: dict, info: str) -> bool:
            if not iscycle(thing, ndict, info):
                res = False
                if isinstance(thing, list):
                    idx = 0
                    for i in thing:
                        res |= recurs(i, ndict, "%s[%d]" % (info, idx))
                        idx += 1
                elif isinstance(thing, Node):
                    res |= thing.check(ndict, info)
                elif isinstance(thing, dict):
                    for k, v in thing.items():
                        res |= recurs(v, ndict, "%s[%s]" % (info, k))
                return res
            return True
        # add ME FIRST
        if len(ndict) == 0:
            ndict['self'] = id(self)
            info = 'self'
        if not iscycle(self, ndict, info):
            res = False
            if len(self) > 0:
                keys = list(self.keys())
                for k in keys:
                    ndict["['" + k + "']"] = id(self[k])
                    res |= recurs(self[k], ndict, "%s[%s]" % (info, k))
            keys = list(vars(self).keys())
            for k in keys:
                ndict["." + k] = id(getattr(self, k))
                res |= recurs(getattr(self, k), ndict, "%s.%s" % (info, k))
            return res
        return True

    def clean(self):
        if len(self) > 0:
            keys = list(self.keys())
            for k in keys:
                del self[k]
        keys = list(vars(self).keys())
        for k in keys:
            delattr(self, k)

    def set(self, othernode):
        """allow to completly mutate the node into any subclasses of Node"""
        self.__class__ = othernode.__class__
        self.clean()
        if len(othernode) > 0:
            for k, v in othernode.items():
                self[k] = v
        for k, v in vars(othernode).items():
            setattr(self, k, v)



# inference mechanisms
from pyrser.type_system import *
from pyrser.error import *
from pyrser.passes.to_yml import *


class Inference:
    # TO REDEFINE IN AST
    def type_algos(self) -> ('self.infer_algo',
                             'self.feedback_algo', ['args']):
        """
        Sub class must return a Tuple of 3 elements:
            - the method to use to infer type.
            - the list of params to used when infer a type.
            - the method to use when feedback a type.

        This is useful to connect AST members to generic algo or
        to overload some specific semantic for your language.
        """
        raise Exception(("%s must inherit from Inference and define" +
                        " type_algos(self) method to support" +
                        " Pyrser Type Systems.") % type(self).__name__)

    def infer_type(self, diagnostic=None):
        # get algo
        type_algo = self.type_algos()
        type_algo[0](type_algo[1], diagnostic)

    def feedback(self, diagnostic=None):
        # get algo
        type_algo = self.type_algos()
        type_algo[2](diagnostic)

    ## INFER ALGOS

    def infer_block(self, body, diagnostic=None):
        """
        Infer type on block is to type each of is sub-element
        """
        # create root type_node for RootBlock
        if not hasattr(self, 'type_node'):
            self.type_node = Scope(is_namespace=False)
        #diagnostic.notify(Severity.INFO, "Infer Block", self.info)
        for e in body:
            e.type_node = Scope(is_namespace=False)
            # linked scope
            e.type_node.set_parent(self.type_node)
            e.infer_type(diagnostic)

    def infer_subexpr(self, expr, diagnostic=None):
        """
        Infer type on the subexpr
        """
        #diagnostic.notify(Severity.INFO, "Infer SubExpr", self.info)
        expr.type_node = Scope(is_namespace=False)
        # linked scope
        expr.type_node.set_parent(self.type_node)
        expr.infer_type(diagnostic)
        self.type_node.clear()
        self.type_node.update(expr.type_node)

    def infer_fun(self, args, diagnostic=None):
        call_expr, arguments = args
        diagnostic.notify(Severity.INFO, "Infer Function call '%s'" % call_expr.value, self.info)
        # 1 - fetch all possible types for the call expr
        call_expr.type_node = Scope(is_namespace=False)
        # linked scope
        call_expr.type_node.set_parent(self.type_node)
        call_expr.infer_type(diagnostic=diagnostic)
        f = call_expr.type_node
        # 2 - fetch all possible types for each parameters
        tparams = []
        i = 0
        for p in arguments:
            p.type_node = Scope(is_namespace=False)
            # linked scope
            p.type_node.set_parent(self.type_node)
            p.infer_type(diagnostic=diagnostic)
            i += 1
            tparams.append(p.type_node)
        # 3 - select overloads
        (final_call_expr, final_tparams) = f.get_by_params(tparams)
        # 4 - record overloads
        self.type_node.clear()
        self.type_node.update(final_call_expr)
        nversion = len(final_call_expr)
        if nversion > 1:
            # too many choice
            self.type_node.need_feedback = True
            return
        elif nversion == 0:
            # TODO: Try type reconstruction if final_tparams[x][0].is_polymorphic
            # ...
            # type error
            details = "\ndetails:\n"
            details += "overloads:\n"
            for fun in f.values():
                details += str(fun.get_compute_sig()) + "\n"
            details += "parameters:\n"
            i = 0
            # TODO: change by final_tparams
            for p in tparams:
                details += "- param[%d] type " % i
                ptypes = []
                for alt in p.values():
                    ptypes.append(alt.tret)
                details += '|'.join(ptypes)+ "\n"
                i += 1
            # TODO: when f are empty
            diagnostic.notify(
                Severity.ERROR,
                "can't match overloads with parameters for function '%s'" % str(f.first().name),
                self.info,
                details
            )
            return
        # here len(self.type_node) == 1 && len(final_tparams) == 1
        # 5 - handle polymorphism
        # TODO: INSIDE A CLASS?
        my_map = dict()
        my_type = self.type_node.first()
        if my_type.tret.is_polymorphic:
            diagnostic.notify(Severity.INFO, "polymorphic return type %s" % str(my_type.get_compute_sig()), self.info)
            # if tret is polymorphic, take type from call_expr if unique, else type come from parameter resolution
            sig = call_expr.type_node.first()
            if len(call_expr.type_node) == 1 and not sig.tret.is_polymorphic:
                diagnostic.notify(Severity.INFO, "call_expr have real type '%s'" % sig.tret, self.info)
                my_map.update(sig.resolution)
                my_type.set_resolved_name(
                    sig.resolution,
                    my_type.tret,
                    sig.tret
                )
        arity = len(my_type.tparams)
        for i in range(arity):
            p = my_type.tparams[i]
            # use AST Injector
            if final_tparams[0][i].first()._translate_to is not None:
                t = final_tparams[0][i].first()._translate_to
                old = arguments[i]
                n = t.notify
                diagnostic.notify(n.severity, n.msg, old.info, details=n.details)
                arguments[i] = self.type_node.callInjector(old, t)
                scope = Scope(is_namespace=False)
                scope.need_feedback = False
                scope.add(t.fun)
                scope.set_parent(self.type_node)
                old.type_node.set_parent(scope)
                arguments[i].type_node = scope
            if p.is_polymorphic:
                # take type in final_tparams
                diagnostic.notify(Severity.INFO, "- param[%d] polymorphic type" % i, arguments[i].info)
                # here len(final_tparams) == 1, only one set for parameters
                sig = final_tparams[0][i].first()
                if not sig.tret.is_polymorphic:
                    diagnostic.notify(Severity.INFO, "- argument[%d] real type '%s'" % (i, sig.tret), arguments[i].info)
                    my_map[p.value] = sig.resolution[sig.tret.value]
                    my_type.set_resolved_name(sig.resolution, p, sig.tret)
        diagnostic.notify(Severity.INFO, "after resolution %s" % str(my_type.get_compute_sig()), self.info)
        # 6 - feedback
        self.map_type = my_map
        self.final_type = my_type.tret
        self.feedback(diagnostic)
        # 7 - Are we finish? Global type reconstruction

    def infer_id(self, ident, diagnostic=None):
        """
        Infer type from an ID!
        - check if ID is declarated in the scope
        - if no ID is polymorphic type
        """
        # check if ID is declared
        defined = self.type_node.get_by_symbol_name(ident)
        if len(defined) > 0:
            # set from matchings declarations
            self.type_node.update(defined)
        else:
            diagnostic.notify(Severity.ERROR, "%s never declared" % self.value, self.info)

    #def infer_id_self_type(self, ident, diagnostic=None):
    #    """
    #    Infer type from an ID!
    #    - check if ID is declarated in the scope
    #    - if no ID is polymorphic type
    #    """
    #    # check if ID is declared
    #    defined = self.type_node.get_by_symbol_name(ident)
    #    if len(defined) > 0:
    #        # set from matchings declarations
    #        self.type_node.update(defined)
    #    else:
    #        # TODO: est de type polymorphique local ... pas tout le temps ?1
    #        # a faire dans une declaration ... ou a l'affectation
    #        # mais pas la
    #        self.type_node.add(Var(ident, '?1'))
    #        self.type_node.need_feedback = True

    def infer_literal(self, args, diagnostic=None):
        """
        Infer type from an LITERAL!
        Type of literal depend of language.
        We adopt a basic convention
        """
        literal, t = args
        self.type_node.add(EvalCtx.from_sig(Val(literal, t)))

    #def infer_operator(self, op, diagnostic=None):
    #    """
    #    Infer type of OPERATOR!
    #    Classic (?1, ?1) -> ?1
    #    """
    #    # by default all operator are polymorphic
    #    # TODO: search proto from scope and need operators in scope
    #    self.type_node.add(Fun(op, '?1', ['?1', '?1']))
    #    self.type_node.need_feedback = True

    ## FEEDBACK ALGOS

    def feedback_block(self, diagnostic=None):
        # normally nothing to do!?!?!
        type_algo = self.type_algos()
        # TODO: body?
        for e in type_algo[1]:
            if e.type_node.need_feedback:
                e.type_node.feedback(diagnostic)

    def feedback_subexpr(self, diagnostic=None):
        self.type_node = self.type_node.get_by_return_type(self.final_type)
        self.type_node.need_feedback = False
        if expr.type_node.need_feedback:
            expr.type_node.feedback(diagnostic)

    def feedback_leaf(self, diagnostic=None):
        self.type_node = self.type_node.get_by_return_type(self.final_type)
        self.type_node.need_feedback = False

    def feedback_fun(self, diagnostic=None):
        self.type_node = self.type_node.get_by_return_type(self.final_type)
        self.type_node.need_feedback = False
        arguments = self.type_node.first().tparams
        nargs = len(arguments)
        type_algo = self.type_algos()
        call_expr, args = type_algo[1]
        if call_expr.type_node.need_feedback:
            diagnostic.notify(Severity.INFO, "feed back call_expr = %s" % self.final_type, call_expr.info)
            call_expr.map_type = self.map_type
            call_expr.final_type = self.final_type
            call_expr.feedback()
        for i in range(nargs):
            p = arguments[i]
            if args[i].type_node.need_feedback:
                diagnostic.notify(Severity.INFO, "feed back p[%d] = %s" % (i, p), args[i].info)
                diagnostic.notify(Severity.INFO, "feed map %s" % self.map_type, args[i].info)
                args[i].map_type = self.map_type
                args[i].final_type = p
                args[i].feedback(diagnostic)

    def feedback_id(self, diagnostic=None):
        # instancie META!!!
        if len(self.type_node) > 1:
            self.type_node = self.type_node.get_by_return_type(self.final_type)
            if len(self.type_node) != 1:
                # ERROR TYPE !!!?!?
                diagnostic.notify(Severity.ERROR, "Type error: too many candidates %s" % str(self.type_node), self.info)
        else:
            the_sig = list(self.type_node.values())[0]
            if the_sig.tret.is_polymorphic:
                self.type_node = EvalCtx.from_sig(the_sig)
                self.type_node.set_resolved_name(self.map_type, the_sig.tret,
                                                 self.final_type)
                # TODO: Why this fucking line... I forgot
                #self.type_node = self.type_node.get_compute_sig()



from pyrser import meta
from pyrser.parsing import Node
from pyrser.parsing.base import BasicParser


@meta.hook(BasicParser, "echo")
def echo_nodes(self, *rest):
    """
    Print nodes.

    example::

        R = [
            In : node #echo("coucou", 12, node)
        ]

    """
    txt = ""
    for thing in rest:
        if isinstance(thing, Node):
            txt += self.value(thing)
        else:
            txt += str(thing)
    print(txt)
    return True



import ast

from pyrser import meta
from pyrser import parsing


class RuleVisitor(ast.NodeVisitor):
    def __init__(self):
        self.in_optional, self.in_loop, self.in_try = 0, 0, 0

    def generic_visit(self, node):
        raise TypeError("Unhandled {} node".format(node.__class__.__name__))

    def __exit_scope(self) -> ast.stmt:
        """Create the appropriate scope exiting statement.

        The documentation only shows one level and always uses
        'return False' in examples.

        'raise AltFalse()' within a try.
        'break' within a loop.
        'return False' otherwise.
        """
        if self.in_optional:
            return ast.Pass()
        if self.in_try:
            return ast.Raise(
                ast.Call(ast.Name('AltFalse', ast.Load()), [], [], None, None),
                None)
        if self.in_loop:
            return ast.Break()
        return ast.Return(ast.Name('False', ast.Load()))

    #TODO(bps): find a better name to describe what this does
    def _clause(self, pt: parsing.ParserTree) -> [ast.stmt]:
        """Normalize a test expression into a statements list.

        Statements list are returned as-is.
        Expression is packaged as:
        if not expr:
            return False
        """
        if isinstance(pt, list):
            return pt
        return [ast.If(ast.UnaryOp(ast.Not(), pt),
                       [self.__exit_scope()],
                       [])]

    def visit_Call(self, node: parsing.Call) -> ast.expr:
        """Generates python code calling the function.

        fn(*args)
        """
        return ast.Call(
            ast.Attribute(
                ast.Name('self', ast.Load),
                node.callObject.__name__,
                ast.Load()),
            [ast.Str(param) for param in node.params],
            [],
            None,
            None)

    def visit_CallTrue(self, node: parsing.CallTrue) -> ast.expr:
        """Generates python code calling the function and returning True.

        lambda: fn(*args) or True
        """
        return ast.Lambda(
            ast.arguments([], None, None, [], None, None, [], []),
            ast.BoolOp(
                ast.Or(),
                [
                    self.visit_Call(node),
                    ast.Name('True', ast.Load())]))

    def visit_Hook(self, node: parsing.Hook) -> ast.expr:
        """Generates python code calling a hook.

        self.evalHook('hookname', self.ruleNodes[-1])
        """
        return ast.Call(
            ast.Attribute(
                ast.Name('self', ast.Load()), 'evalHook', ast.Load()),
            [
                ast.Str(node.name),
                ast.Subscript(
                    ast.Attribute(
                        ast.Name('self', ast.Load()), 'ruleNodes', ast.Load()),
                    ast.Index(ast.UnaryOp(ast.USub(), ast.Num(1))),
                    ast.Load())],
            [],
            None,
            None)

    def visit_Rule(self, node: parsing.Rule) -> ast.expr:
        """Generates python code calling a rule.

        self.evalRule('rulename')
        """
        return ast.Call(
            ast.Attribute(ast.Name('self', ast.Load()),
                          'evalRule', ast.Load()),
            [ast.Str(node.name)], [], None, None)

    def visit_Capture(self, node: parsing.Capture) -> [ast.stmt] or ast.expr:
        """Generates python code to capture text consumed by a clause.

        #If all clauses can be inlined
        self.beginTag('tagname') and clause and self.endTag('tagname')

        if not self.beginTag('tagname'):
            return False
        <code for the clause>
        if not self.endTag('tagname'):
            return False
        """
        begintag = ast.Attribute(
            ast.Name('self', ast.Load()), 'beginTag', ast.Load())
        endtag = ast.Attribute(
            ast.Name('self', ast.Load()), 'endTag', ast.Load())
        begin = ast.Call(begintag, [ast.Str(node.tagname)], [], None, None)
        end = ast.Call(endtag, [ast.Str(node.tagname)], [], None, None)
        result = [begin, self.visit(node.pt), end]
        for clause in result:
            if not isinstance(clause, ast.expr):
                break
        else:
            return ast.BoolOp(ast.And(), result)
        res = []
        for stmt in map(self._clause, result):
            res.extend(stmt)
        return res

    def visit_Scope(self, node: parsing.Capture) -> [ast.stmt] or ast.expr:
        """Generates python code for a scope.

        if not self.begin():
            return False
        res = self.pt()
        if not self.end():
            return False
        return res
        """
        return ast.Name('scope_not_implemented', ast.Load())
        raise NotImplementedError()

    #TODO(bps): forbid nested alt
    def visit_Alt(self, node: parsing.Alt) -> [ast.stmt]:
        """Generates python code for alternatives.

        try:
            try:
                <code for clause>  #raise AltFalse when alternative is False
                raise AltTrue()
            except AltFalse:
                pass
            return False
        except AltTrue:
            pass
        """
        clauses = [self.visit(clause) for clause in node.ptlist]
        for clause in clauses:
            if not isinstance(clause, ast.expr):
                break
        else:
            return ast.BoolOp(ast.Or(), clauses)
        res = ast.Try([], [ast.ExceptHandler(
            ast.Name('AltTrue', ast.Load()), None, [ast.Pass()])], [], [])
        alt_true = [ast.Raise(ast.Call(
            ast.Name('AltTrue', ast.Load()), [], [], None, None), None)]
        alt_false = [ast.ExceptHandler(
            ast.Name('AltFalse', ast.Load()), None, [ast.Pass()])]
        self.in_try += 1
        for clause in node.ptlist:
            res.body.append(
                ast.Try(self._clause(self.visit(clause)) + alt_true,
                        alt_false, [], []))
        self.in_try -= 1
        res.body.append(self.__exit_scope())
        return [res]

    def combine_exprs_for_clauses(self, exprs: [ast.expr]) -> [ast.stmt]:
        expr = exprs[0] if len(exprs) == 1 else ast.BoolOp(ast.And(), exprs)
        return self._clause(expr)

    def visit_Seq(self, node: parsing.Seq) -> [ast.stmt] or ast.expr:
        """Generates python code for clauses.

        #Continuous clauses which can can be inlined are combined with and
        clause and clause

        if not clause:
            return False
        if not clause:
            return False
        """
        exprs, stmts = [], []
        for clause in node.ptlist:
            clause_ast = self.visit(clause)
            if isinstance(clause_ast, ast.expr):
                exprs.append(clause_ast)
            else:
                if exprs:
                    stmts.extend(self.combine_exprs_for_clauses(exprs))
                    exprs = []
                stmts.extend(self._clause(clause_ast))
        if not stmts:
            return ast.BoolOp(ast.And(), exprs)
        if exprs:
            stmts.extend(self.combine_exprs_for_clauses(exprs))
        return stmts

    def visit_RepOptional(self, node: parsing.RepOptional) -> ([ast.stmt] or
                                                               ast.expr):
        """Generates python code for an optional clause.

        <code for the clause>
        """
        cl_ast = self.visit(node.pt)
        if isinstance(cl_ast, ast.expr):
            return ast.BoolOp(ast.Or(), [cl_ast, ast.Name('True', ast.Load())])
        self.in_optional += 1
        cl_ast = self.visit(node.pt)
        self.in_optional -= 1
        return cl_ast

    def visit_Rep0N(self, node: parsing.Rep0N) -> [ast.stmt]:
        """Generates python code for a clause repeated 0 or more times.

        #If all clauses can be inlined
        while clause:
            pass

        while True:
            <code for the clause>
        """
        cl_ast = self.visit(node.pt)
        if isinstance(cl_ast, ast.expr):
            return [ast.While(cl_ast, [ast.Pass()], [])]
        self.in_loop += 1
        clause = self._clause(self.visit(node.pt))
        self.in_loop -= 1
        return [ast.While(ast.Name('True', ast.Load()), clause, [])]

    def visit_Rep1N(self, node: parsing.Rep0N) -> [ast.stmt]:
        """Generates python code for a clause repeated 1 or more times.

        <code for the clause>
        while True:
            <code for the clause>
        """
        clause = self.visit(node.pt)
        if isinstance(clause, ast.expr):
            return (self._clause(clause) + self.visit_Rep0N(node))
        self.in_loop += 1
        clause = self._clause(self.visit(node.pt))
        self.in_loop -= 1
        return self._clause(self.visit(node.pt)) + [
            ast.While(ast.Name('True', ast.Load()), clause, [])]


def rule_topython(rule: parsing.ParserTree) -> [ast.stmt]:
    return RuleVisitor().visit(rule)


def parserrule_topython(parser: parsing.BasicParser,
                        rulename: str) -> ast.FunctionDef:
    """Generates code for a rule.

    def rulename(self):
        <code for the rule>
        return True
    """
    visitor = RuleVisitor()
    rule = parser._rules[rulename]
    fn_args = ast.arguments([ast.arg('self', None)], None, None, [], None,
                            None, [], [])
    body = visitor._clause(rule_topython(rule))
    body.append(ast.Return(ast.Name('True', ast.Load())))
    return ast.FunctionDef(rulename, fn_args, body, [], None)



from pyrser import meta
from pyrser.parsing import Node
from pyrser.parsing.base import BasicParser


@meta.hook(BasicParser, "vars")
def vars_nodes(self, rest):
    """
    Vars one node instance.

    example::

        R = [
            In : node #vars(node)
        ]

    """
    print(vars(rest))
    return True



from pyrser import meta
from pyrser.parsing.base import BasicParser


@meta.hook(BasicParser, "dump_nodes")
def dump_nodes(self):
    """
    Dump tag,rule,id and value cache. For debug.

    example::

        R = [
            #dump_nodes
        ]

    """
    print("DUMP NODE LOCAL INFOS")
    try:
        print("map Id->node name")
        for k, v in self.id_cache.items():
            print("[%d]=%s" % (k, v))
        print("map tag->capture infos")
        for k, v in self.tag_cache.items():
            print("[%s]=%s" % (k, v))
        print("map nodes->tag resolution")
        for k, v in self.rule_nodes.items():
            txt = "['%s']=%d" % (k, id(v))
            if k in self.tag_cache:
                tag = self.tag_cache[k]
                txt += " tag <%s>" % tag
                k = "%d:%d" % (tag._begin, tag._end)
                if k in self._stream.value_cache:
                    txt += " cache <%s>" % self._stream.value_cache[k]
            print(txt)
    except Exception as err:
        print("RECV Exception %s" % err)
    import sys
    sys.stdout.flush()
    return True



#!/usr/bin/env python3.3

import setuptools

version = '0.0'
release = '0.0.3'

setuptools.setup(
    name='pyrser',
    version=release,
    url='https://code.google.com/p/pyrser/',
    license='GPLv3',
    maintainer='Lionel Auroux',
    maintainer_email='lionel.auroux@gmail.com',
    keywords="parse parser parsing peg packrat grammar language",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'License :: OSI Approved ::' +
            ' GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Compilers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: General',
    ],
    packages=['pyrser', 'pyrser.directives', 'pyrser.hooks', 'pyrser.parsing',
              'pyrser.passes', 'pyrser.type_system']
    #    test_loader='unittest:TestLoader',
    #    test_suite='tests'
)



from collections import *
from pyrser import fmt
from pyrser.type_system.fun import *
from pyrser.error import *


# for type translation
class Translator:
    def __init__(self, fun: Fun, notify: Notification=None):
        if not isinstance(fun, Fun):
            raise TypeError("1st parameter must be a type_system.Fun")
        if not isinstance(notify, Notification):
            raise TypeError("2nd parameter must be a error.Notification")
        if fun.arity < 1:
            raise TypeError("type_system.Fun in 1st parameter must have an arity >= 1 (here %d)" % fun.arity)
        self._type_source = fun.this_type
        self._type_target = fun.return_type
        self._fun = fun
        self._notify = notify

    def __hash__(self) -> str:
        return self._type_target.__hash__()

    @property
    def source(self) -> str:
        return self._type_source

    @property
    def target(self) -> str:
        return self._type_target

    @property
    def fun(self) -> Fun:
        return self._fun

    @property
    def notify(self) -> Notification:
        return self._notify

    def to_fmt(self, with_from=False) -> fmt.indentable:
        """
        Return a Fmt representation of Translator for pretty-printing
        """
        txt = fmt.sep("\n", [
            fmt.sep(" ", [self._type_source, "to", self._type_target, '=', self._fun.to_fmt()]),
            self._notify.get_content(with_from)
        ])
        return txt

    def __repr__(self) -> str:
        return str(self.to_fmt())


class MapTargetTranslate(dict):
    def __init__(self):
        self._internal = {}
        self._type_source = None

    @property
    def key(self) -> str:
        return self._type_source

    def to_fmt(self, with_from=False) -> fmt.indentable:
        txt = fmt.block('{\n', '\n}', [])
        items = fmt.sep('\n---\n', [])
        for k in sorted(self._internal.keys()):
            items.lsdata.append(fmt.sep(': ', [k, fmt.tab([self._internal[k].to_fmt(with_from)])]))
        txt.lsdata.append(fmt.tab([items]))
        return txt

    def __repr__(self) -> str:
        return str(self.to_fmt())

    def __len__(self) -> int:
        return len(self._internal)

    def __delitem__(self, key: str):
        del self._internal[key]

    def __contains__(self, key: str) -> bool:
        return key in self._internal

    def __iter__(self) -> iter:
        return iter(self._internal)

    def __reversed(self) -> reversed:
        return reversed(self._internal)

    def __getitem__(self, key: str) -> Translator:
        return self._internal[key]

    def __setitem__(self, key: str, val: Translator):
        if not isinstance(val, Translator):
            raise TypeError("value must be a Translator")
        if key != val.target:
            raise KeyError("Key:%s != Translator:%s, bad key of Translator" % (key, val.target))
        if self._type_source is None:
            self._type_source = val.source
        elif self._type_source != val.source:
            raise KeyError("Map.source:%s != Translator.source:%s, Translator incompatible with map" % (self._type_source, val._type_source))
        self._internal[key] = val


# forward just for annotation (not the same id that final type)
class MapSourceTranslate:
    pass


class MapSourceTranslate(dict):
    def __init__(self):
        self._internal = ChainMap()

    def set_parent(self, parent: MapSourceTranslate):
        tmp = self._internal
        self._internal = parent._internal.new_child()
        self._internal.maps[0].update(tmp)

    def to_fmt(self, with_from=False) -> fmt.indentable:
        txt = fmt.block('{\n', '\n}', [])
        items = fmt.sep('\n---\n', [])
        for k in sorted(self._internal.keys()):
            items.lsdata.append(fmt.sep(': ', [k, fmt.tab([self._internal[k].to_fmt(with_from)])]))
        txt.lsdata.append(fmt.tab([items]))
        return txt

    def __repr__(self) -> fmt.indentable:
        return str(self.to_fmt())

    def __len__(self) -> int:
        return len(self._internal)

    def __delitem__(self, key: str):
        del self._internal[key]

    def __contains__(self, key: str or tuple) -> bool:
        if isinstance(key, str):
            return key in self._internal
        if isinstance(key, tuple):
            return (key[0] in self._internal
                    and key[1] in self._internal[key[0]]
                    )
        raise TypeError("parameter must be str or tuple")

    def __iter__(self) -> iter:
        return iter(self._internal)

    def __reversed(self) -> reversed:
        return reversed(self._internal)

    def addTranslator(self, val: Translator, as_global=False):
        if not isinstance(val, Translator):
            raise TypeError("value must be a Translator")
        t1 = val.source
        t2 = val.target
        current_map = self._internal
        if as_global:
            current_map = self._internal.maps[-1]
        if t1 not in current_map:
            current_map[t1] = MapTargetTranslate()
        if t2 in current_map[t1]:
            raise KeyError("Mapping %s to %s already exist" % (t1, t2))
        current_map[t1][t2] = val

    def __getitem__(self, key: str) -> MapTargetTranslate:
        if key not in self._internal:
            self._internal[key] = MapTargetTranslate()
        return self._internal[key]

    def __setitem__(self, key: str, val: MapTargetTranslate):
        if not isinstance(val, MapTargetTranslate):
            raise TypeError("value must be a MapTargetTranslate")
        if key != val.key:
            raise KeyError("Key:%s != Translator:%s, bad key of Translator" % (key, val.key))
        if val._type_source is None:
            val._type_source = key
        elif val._type_source != key:
            raise KeyError("MapSource.key:%s != MapTarget.source:%s, MapTarget incompatible with key" % (key, val._type_source))
        self._internal[key] = val



# scope for type checking
import weakref
from collections import *
from pyrser import fmt, meta
from pyrser.type_system.symbol import *
from pyrser.type_system.signature import *
from pyrser.type_system.evalctx import *
from pyrser.type_system.translator import *
from pyrser.parsing.node import *
from pyrser.passes.to_yml import *


StateScope = meta.enum('FREE', 'LINKED', 'EMBEDDED')

# forward just for annotation (not the same id that final type)
class Scope:
    pass


class Scope(Symbol):
    """
    Scope of Signature for a Scope/namespace/type etc...
    Basic abstraction of type checking.
    Scope is not a 'pure' python set but something between a set and a dict...
    """

    def __init__(self, name: str=None, sig: [Signature]=None,
                 state=StateScope.FREE, is_namespace=True):
        """Unnamed scope for global scope
        
        A Scope have basically 3 possibles states:
        FREE: it's a standalone Scope, generally the global Scope
        LINKED: the Scope is related to another Scope for type resolution, like compound statement
        EMBEDDED: the Scope was added into another Scope, it forwards all 'in' calls...like namespacing

        A Scope could or couldn't act like a namespace.
        All Signature added into a 'namespaced' Scope was prefixed by the Scope name.
        """
        if name is not None and not isinstance(name, str):
            raise TypeError("name must be a str object.")
        super().__init__(name)
        # during typing, this scope need or not feedback pass
        self.need_feedback = False
        # state of the scope
        self.state = state
        self.is_namespace = is_namespace
        # internal mapping for Type Conversion
        self.mapTypeTranslate = MapSourceTranslate()
        # AST Translator Injector use to add Translator in AST
        # def astTranslatorInjector(old: Node, trans: Translator, d: Diagnostic, li: LocationInfo) -> Node
        self.astTranslatorInjector = None
        # TODO: ...could be unusable
        self._ntypes = 0
        self._nvars = 0
        self._nfuns = 0
        # internal store of Signature
        self._hsig = {}
        if sig is not None:
            if isinstance(sig, Signature) or isinstance(sig, Scope):
                self.add(sig)
            elif len(sig) > 0:
                self.update(sig)

    def set_parent(self, parent: Scope) -> object:
        Symbol.set_parent(self, parent)
        if parent is not None:
            self.mapTypeTranslate.set_parent(parent.mapTypeTranslate)
        if hasattr(self, 'state') and self.state == StateScope.FREE:
            self.state = StateScope.LINKED
        return self

    def set_name(self, name: str):
        """
        You could set the name after construction
        """
        self.name = name
        # update internal names
        lsig = self._hsig.values()
        self._hsig = {}
        for s in lsig:
            self._hsig[s.internal_name()] = s

    def count_types(self) -> int:
        """
        Count subtypes
        """
        n = 0
        for s in self._hsig.values():
            if type(s).__name__ == 'Type':
                n += 1
        return n

    def count_vars(self) -> int:
        """
        Count var define by this scope
        """
        n = 0
        for s in self._hsig.values():
            if hasattr(s, 'is_var') and s.is_var():
                n += 1
        return n

    def count_funs(self) -> int:
        """
        Count function define by this scope
        """
        n = 0
        for s in self._hsig.values():
            if hasattr(s, 'is_fun') and s.is_fun():
                n += 1
        return n

    def __update_count(self):
        """
        Update internal counters
        """
        self._ntypes = self.count_types()
        self._nvars = self.count_vars()
        self._nfuns = self.count_funs()

    # ======== PRETTY PRINTING ========
    def to_fmt(self) -> fmt.indentable:
        """
        Return an Fmt representation for pretty-printing
        """
        qual = "scope"
        txt = fmt.sep(" ", [qual])
        name = self.show_name()
        if name != "":
            txt.lsdata.append(name)
        if len(self._hsig) > 0 or len(self.mapTypeTranslate) > 0:
            lsb = []
            if len(self.mapTypeTranslate) > 0:
                lsb.append("translate:\n")
                lsb.append(fmt.end("\n", self.mapTypeTranslate.to_fmt()))
            for k in sorted(self._hsig.keys()):
                s = self._hsig[k]
                lsb.append(fmt.end("\n", [s.to_fmt()]))
            block = fmt.block(":\n", "", fmt.tab(lsb))
            txt.lsdata.append(block)
        return txt

    def __repr__(self) -> str:
        """
        Internal representation
        """
        return repr(self._hsig)

    def __str__(self) -> str:
        """
        Usefull representation
        """
        return str(self.to_fmt())

    def __getstate__(self):
        """
        For pickle don't handle weakrefs...
        """
        state = self.__dict__.copy()
        del state['parent']
        return state

    ## TODO: __getitem__, __setitem__, __delete__??

    # ======== SET OPERATORS OVERLOADING ========
    # in
    def __contains__(self, s: Signature) -> bool:
        """
        check if a Signature or a Type is declared in a Scope
        """
        # fail if in global
        from pyrser.type_system.type import Type
        found = False
        txt = ""
        if isinstance(s, Signature):
            txt = s.internal_name()
        elif isinstance(s, Type):
            txt = s.type_name
        elif isinstance(s, str):
            txt = s
        found = (txt in self._hsig)
        if not found and self.parent is not None:
            if self.state != StateScope.FREE and isinstance(s, Type):
                found = (s.type_name in self.parent())
            if self.state == StateScope.EMBEDDED:
                found = (txt in self.parent())
        return found

    # |=
    def __ior__(self, sig: list or Scope) -> Scope:
        """|= operator"""
        return self.update(sig)

    def update(self, sig: list or Scope) -> Scope:
        """Update the Set with values of another Set"""
        values = sig
        if hasattr(sig, 'values'):
            values = sig.values()
        for s in values:
            if self.is_namespace:
                s.set_parent(self)
            if isinstance(s, Scope):
                s.state = StateScope.EMBEDDED
            self._hsig[s.internal_name()] = s
        self.__update_count()
        return self

    # |
    def __or__(self, sig: Scope) -> Scope:
        """| operator"""
        return self.union(sig)

    def union(self, sig: Scope) -> Scope:
        """Create a new Set produce by the union of 2 Set"""
        new = Scope(sig=self._hsig.values(), state=self.state)
        new |= sig
        return new

    # &=
    def __iand__(self, sig: Scope) -> Scope:
        """&= operator"""
        return self.intersection_update(sig)

    def intersection_update(self, oset: Scope) -> Scope:
        """Update Set with common values of another Set"""
        keys = list(self._hsig.keys())
        for k in keys:
            if k not in oset:
                del self._hsig[k]
            else:
                self._hsig[k] = oset.get(k)
        return self

    # &
    def __and__(self, sig: Scope) -> Scope:
        """& operator"""
        return self.intersection(sig)

    def intersection(self, sig: Scope) -> Scope:
        """Create a new Set produce by the intersection of 2 Set"""
        new = Scope(sig=self._hsig.values(), state=self.state)
        new &= sig
        return new

    # -=
    def __isub__(self, sig: Scope) -> Scope:
        """-= operator"""
        return self.difference_update(sig)

    def difference_update(self, oset: Scope) -> Scope:
        """Remove values common with another Set"""
        keys = list(self._hsig.keys())
        for k in keys:
            if k in oset:
                del self._hsig[k]
        return self

    # -
    def __sub__(self, sig: Scope) -> Scope:
        """- operator"""
        return self.difference(sig)

    def difference(self, sig: Scope) -> Scope:
        """Create a new Set produce by a Set subtracted by another Set"""
        new = Scope(sig=self._hsig.values(), state=self.state)
        new -= sig
        return new

    # ^=
    def __ixor__(self, sig: Scope) -> Scope:
        """^= operator"""
        return self.symmetric_difference_update(sig)

    def symmetric_difference_update(self, oset: Scope) -> Scope:
        """Remove common values and Update specific values from another Set"""
        skey = set()
        keys = list(self._hsig.keys())
        for k in keys:
            if k in oset:
                skey.add(k)
        for k in oset._hsig.keys():
            if k not in skey:
                self._hsig[k] = oset.get(k)
        for k in skey:
            del self._hsig[k]
        return self

    # ^
    def __xor__(self, sig: Scope) -> Scope:
        """^ operator"""
        return self.symmetric_difference(sig)

    def symmetric_difference(self, sig: Scope) -> Scope:
        """Create a new Set with values present in only one Set"""
        new = Scope(sig=self._hsig.values(), state=self.state)
        new ^= sig
        return new

    # ======== SCOPE MANIPULATION ========
    def add(self, it: Signature) -> bool:
        """
        Add it to the Set
        """
        if isinstance(it, Scope):
            it.state = StateScope.EMBEDDED
        txt = it.internal_name()
        it.set_parent(self)
        if self.is_namespace:
            txt =  it.internal_name()
        if txt == "":
            txt = '_' + str(len(self._hsig))
        if txt in self._hsig:
            raise KeyError("Already exists %s" % txt)
        self._hsig[txt] = it
        self.__update_count()
        return True

    def remove(self, it: Signature) -> bool:
        """
        Remove it but raise KeyError if not found
        """
        txt = it.internal_name()
        if txt not in self._hsig:
            raise KeyError(it.show_name() + ' not in Set')
        sig = self._hsig[txt]
        if isinstance(sig, Scope):
            sig.state = StateScope.LINKED
        del self._hsig[txt]
        return True

    def discard(self, it: Signature) -> bool:
        """
        Remove it only if present
        """
        txt = it.internal_name()
        if txt in self._hsig:
            sig = self._hsig[txt]
            if isinstance(sig, Scope):
                sig.state = StateScope.LINKED
            del self._hsig[txt]
            return True
        return False

    def clear(self) -> bool:
        """
        Clear all signatures in the Set
        """
        self._hsig.clear()
        return True

    def pop(self) -> Signature:
        """
        Pop a random Signature
        """
        return self._hsig.popitem()

    # ======== SCOPE ITERATION ========
    def __len__(self) -> int:
        """
        Len of the Set
        """
        return len(self._hsig)

    def values(self) -> [Signature]:
        """
        Retrieve all values
        """
        if self.state == StateScope.EMBEDDED and self.parent is not None:
            return list(self._hsig.values()) + list(self.parent().values())
        else:
            return self._hsig.values()

    def keys(self) -> [str]:
        """
        Retrieve all keys
        """
        return self._hsig.keys()

    # ======== SCOPE QUERY ========
    def first(self) -> Signature:
        """
        Retrieve the first Signature ordered by mangling descendant
        """
        k = sorted(self._hsig.keys())
        return self._hsig[k[0]]

    def last(self) -> Signature:
        """
        Retrieve the last Signature ordered by mangling descendant
        """
        k = sorted(self._hsig.keys())
        return self._hsig[k[-1]]

    def get(self, key: str, default=None) -> Signature:
        """
        Get a signature instance by its internal_name
        """
        item = default
        if key in self._hsig:
            item = self._hsig[key]
        return item

    def get_by_symbol_name(self, name: str) -> Scope:
        """
        Retrieve a Set of all signature by symbol name
        """
        lst = []
        for s in self.values():
            if s.name == name:
                # create an EvalCtx only when necessary
                lst.append(EvalCtx.from_sig(s))
        # include parent
        # TODO: see all case of local redefinition for
        #       global overloads
        # possible algos... take all with different internal_name
        if len(lst) == 0:
            p = self.get_parent()
            if p is not None:
                return p.get_by_symbol_name(name)
        rscope = Scope(sig=lst, state=StateScope.LINKED, is_namespace=False)
        # inherit type/translation from parent
        rscope.set_parent(self)
        return rscope

    def getsig_by_symbol_name(self, name: str) -> Signature:
        """
        Retrieve the unique Signature of a symbol.
        Fail if the Signature is not unique
        """
        subscope = self.get_by_symbol_name(name)
        if len(subscope) != 1:
            raise KeyError("%s have multiple candidates in scope" % name)
        v = list(subscope.values())
        return v[0]

    def get_by_return_type(self, tname: str) -> Scope:
        """
        Retrieve a Set of all signature by (return) type
        """
        lst = []
        for s in self.values():
            if hasattr(s, 'tret') and s.tret == tname:
                lst.append(EvalCtx.from_sig(s))
        rscope = Scope(sig=lst, state=StateScope.LINKED, is_namespace=False)
        # inherit type/translation from parent
        rscope.set_parent(self)
        return rscope

    def get_all_polymorphic_return(self) -> bool:
        """
        For now, polymorphic return type are handle by symbol artefact.

        --> possible multi-polymorphic but with different constraint attached!
        """
        lst = []
        for s in self.values():
            if hasattr(s, 'tret') and s.tret.is_polymorphic:
                # encapsulate s into a EvalCtx for meta-var resolution
                lst.append(EvalCtx.from_sig(s))
        rscope = Scope(sig=lst, state=StateScope.LINKED, is_namespace=False)
        # inherit type/translation from parent
        rscope.set_parent(self)
        return rscope

    def get_by_params(self, params: [Scope]) -> (Scope, [[Scope]]):
        """
        Retrieve a Set of all signature that match the parameter list.
        Return a pair.
            pair[0] the overloads for the functions
            pair[1] the overloads for the parameters (a list of candidate list of parameters)
        """
        lst = []
        scopep = []
        # for each of our signatures
        for s in self.values():
            # for each params of this signature
            if hasattr(s, 'tparams'):
                # number of matched params
                mcnt = 0
                # temporary collect
                nbparam_sig = len(s.tparams)
                nbparam_candidates = len(params)
                # don't treat signature too short
                if nbparam_sig > nbparam_candidates:
                    continue
                # don't treat call signature too long if not variadic
                if nbparam_candidates > nbparam_sig and not s.variadic:
                    continue
                tmp = [None] * nbparam_candidates
                variadic_types = []
                for i in range(nbparam_candidates):
                    tmp[i] = Scope(state=StateScope.LINKED)
                    tmp[i].set_parent(self)
                    # match param of the expr
                    if i < nbparam_sig:
                        m = params[i].get_by_return_type(s.tparams[i])
                        if len(m) > 0:
                            mcnt += 1
                            tmp[i].update(m)
                        else:
                            # co/contra-variance
                            # we just need to search a t1->t2
                            # and add it into the tree (with/without warnings)
                            t1 = params[i]
                            t2 = s.tparams[i]
                            # if exist a fun (t1) -> t2
                            (is_convertible,
                             signature,
                             translator
                             ) = t1.findTranslationTo(t2)
                            if is_convertible:
                                # add a translator in the EvalCtx
                                signature.use_translator(translator)
                                mcnt += 1
                                nscope = Scope(
                                    sig=[signature],
                                    state=StateScope.LINKED,
                                    is_namespace=False
                                    )
                                nscope.set_parent(self)
                                tmp[i].update(nscope)
                            elif s.tparams[i].is_polymorphic:
                                # handle polymorphic parameter
                                mcnt += 1
                                if not isinstance(params[i], Scope):
                                    raise Exception(
                                        "params[%d] must be a Scope" % i
                                    )
                                tmp[i].update(params[i])
                            else:
                                # handle polymorphic return type
                                m = params[i].get_all_polymorphic_return()
                                if len(m) > 0:
                                    mcnt += 1
                                    tmp[i].update(m)
                    # for variadic extra parameters
                    else:
                        mcnt += 1
                        if not isinstance(params[i], Scope):
                            raise Exception("params[%d] must be a Scope" % i)
                        variadic_types.append(params[i].first().tret)
                        tmp[i].update(params[i])
                # we have match all candidates
                if mcnt == len(params):
                    # select this signature but
                    # box it (with EvalCtx) for type resolution
                    lst.append(EvalCtx.from_sig(s))
                    lastentry = lst[-1]
                    if lastentry.variadic:
                        lastentry.use_variadic_types(variadic_types)
                    scopep.append(tmp)
        rscope = Scope(sig=lst, state=StateScope.LINKED, is_namespace=False)
        # inherit type/translation from parent
        rscope.set_parent(self)
        return (rscope, scopep)

    def addTranslator(self, val: Translator, as_global=False):
        return self.mapTypeTranslate.addTranslator(val, as_global=as_global)

    def addTranslatorInjector(self, ast_method):
        """
        Could be redefine in a subscope
        """
        if self.astTranslatorInjector is not None:
            raise TypeError("Already define a translator injector")
        self.astTranslatorInjector = ast_method

    def callInjector(self, old: Node, trans: Translator) -> Node:
        """
        If don't have injector call from parent
        """
        if self.astTranslatorInjector is None:
            if self.parent is not None:
                # TODO: think if we forward for all StateScope
                # forward to parent scope
                return self.parent().callInjector(old, trans)
            else:
                raise TypeError("Must define an Translator Injector")
        return self.astTranslatorInjector(old, trans)

    def findTranslationTo(self, t2: str) -> (bool, Signature, Translator):
        """
        Find an arrow (->) aka a function able to translate something to t2
        """
        if not t2.is_polymorphic:
            collect = []
            for s in self.values():
                t1 = s.tret
                if t1.is_polymorphic:
                    continue
                if (s.tret in self.mapTypeTranslate):
                    if (t2 in self.mapTypeTranslate[t1]):
                        collect.append((
                            True,
                            s,
                            self.mapTypeTranslate[t1][t2]
                        ))
            # if len > 1 too many candidates
            if len(collect) == 1:
                return collect[0]
        return (False, None, None)
